import logging
from collections.abc import AsyncIterator, Iterator
from typing import Callable, TYPE_CHECKING

from opentelemetry import context as context_api, trace
from opentelemetry.context import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.sdk.trace import _Span
from opentelemetry.trace import (
    SpanKind,
    Status,
    StatusCode,
    get_current_span,
)

from baserun.constants import UNTRACED_SPAN_PARENT_NAME
from baserun.instrumentation.span_attributes import SpanAttributes
from baserun.v1.baserun_pb2 import Run

if TYPE_CHECKING:
    from baserun.instrumentation.base_instrumentor import BaseInstrumentor

logger = logging.getLogger(__name__)


def async_instrumented_wrapper(
    wrapped_fn: Callable, instrumentor: "BaseInstrumentor", span_name: str
):
    """Generates a function (`instrumented_function`) which instruments the original function (`wrapped_fn`)"""

    async def instrumented_function(*args, **kwargs):
        from baserun import Baserun

        if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
            return await wrapped_fn(*args, **kwargs)

        tracer_provider = trace.get_tracer_provider()
        tracer = tracer_provider.get_tracer("baserun")

        parent_span: _Span = get_current_span()
        if not parent_span.is_recording() and not Baserun.current_test_suite:
            run = Baserun.get_or_create_current_run(
                name=UNTRACED_SPAN_PARENT_NAME,
                trace_type=Run.RunType.RUN_TYPE_PRODUCTION,
            )
            parent_span = tracer.start_span(
                UNTRACED_SPAN_PARENT_NAME,
                kind=SpanKind.CLIENT,
                attributes={SpanAttributes.BASERUN_RUN: Baserun.serialize_run(run)},
            )

        span = tracer.start_span(
            **setup_span(span_name=span_name, parent_span=parent_span)
        )
        auto_end_span = True
        try:
            # Activate the span in the current context, but don't end it automatically
            with trace.use_span(span, end_on_exit=False):
                # Capture request attributes
                set_request_attributes(
                    instrumentor=instrumentor, span=span, kwargs=kwargs
                )

                # Actually call the wrapped method
                try:
                    response = await wrapped_fn(*args, **kwargs)

                    span.set_status(Status(StatusCode.OK))
                except Exception as e:
                    span.set_status(
                        Status(
                            description=str(e),
                            status_code=StatusCode.ERROR,
                        )
                    )
                    raise e

                # If this is a streaming response, wrap it, so we can capture each chunk
                if isinstance(response, AsyncIterator):
                    wrapped_response = instrumentor.async_generator_wrapper(
                        response, span
                    )
                    # The span will be ended inside the generator once it's finished
                    auto_end_span = False
                    return wrapped_response

                handle_response(instrumentor, span, response)
        finally:
            if auto_end_span:
                if span.is_recording():
                    span.end()

                if (
                    parent_span.name == UNTRACED_SPAN_PARENT_NAME
                    and parent_span.is_recording()
                    and not Baserun.current_test_suite
                ):
                    parent_span.end()

        return response

    return instrumented_function


def instrumented_wrapper(
    wrapped_fn: Callable, instrumentor: "BaseInstrumentor", span_name: str
):
    """Generates a function (`instrumented_function`) which instruments the original function (`wrapped_fn`)"""

    def instrumented_function(*args, **kwargs):
        """Replacement for the original function (`wrapped_fn`). Will perform instrumentation, call `wrapped_fn`,
        perform more instrumentation, and then return the result of the previous call to `wrapped_fn`.
        """
        from baserun import Baserun

        if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
            return wrapped_fn(*args, **kwargs)

        tracer_provider = trace.get_tracer_provider()
        tracer = tracer_provider.get_tracer("baserun")

        parent_span: _Span = get_current_span()
        # If a call is made outside of a traced function we need to create a parent
        if not parent_span.is_recording() and not Baserun.current_test_suite:
            run = Baserun.get_or_create_current_run(
                name="untraced",
                trace_type=Run.RunType.RUN_TYPE_PRODUCTION,
            )
            parent_span = tracer.start_span(
                UNTRACED_SPAN_PARENT_NAME,
                kind=SpanKind.CLIENT,
                attributes={SpanAttributes.BASERUN_RUN: Baserun.serialize_run(run)},
            )

        span = tracer.start_span(
            **setup_span(span_name=span_name, parent_span=parent_span)
        )

        auto_end_span = True
        try:
            # Activate the span in the current context, but don't end it automatically
            with trace.use_span(span, end_on_exit=False):
                # Capture request attributes
                set_request_attributes(
                    instrumentor=instrumentor, span=span, kwargs=kwargs
                )

                # Actually call the wrapped method
                try:
                    response = wrapped_fn(*args, **kwargs)

                    span.set_status(Status(StatusCode.OK))
                except Exception as e:
                    span.set_status(
                        Status(
                            description=str(e),
                            status_code=StatusCode.ERROR,
                        )
                    )
                    raise e

                # If this is a streaming response, wrap it, so we can capture each chunk
                if isinstance(response, Iterator):
                    wrapped_response = instrumentor.generator_wrapper(response, span)
                    # The span will be ended inside the generator once it's finished
                    auto_end_span = False
                    return wrapped_response

                handle_response(instrumentor, span, response)
        finally:
            if auto_end_span:
                if span.is_recording():
                    span.end()

                if (
                    parent_span.name == UNTRACED_SPAN_PARENT_NAME
                    and parent_span.is_recording()
                    and not Baserun.current_test_suite
                ):
                    parent_span.end()

        return response

    return instrumented_function


def setup_span(span_name: str, parent_span: _Span) -> dict:
    request_type = span_name.split(".")[-1]
    parent_span.set_attribute(SpanAttributes.LLM_REQUEST_TYPE, request_type)

    return {
        "name": f"baserun.{span_name}",
        "kind": SpanKind.CLIENT,
        "attributes": parent_span.attributes,
    }


def set_request_attributes(
    instrumentor: "BaseInstrumentor", span: _Span, kwargs: dict
) -> _Span:
    try:
        instrumentor.set_request_attributes(span, kwargs)
    except Exception as e:
        logger.warning(f"Failed to set input attributes for Baserun span, error: {e}")

    return span


def handle_response(instrumentor: "BaseInstrumentor", span: _Span, response):
    from baserun import Baserun

    run = Baserun.current_run()
    if not run:
        logger.warning("Baserun data not propagated correctly, cannot send data")
        return response

    span.set_attribute(SpanAttributes.BASERUN_RUN, Baserun.serialize_run(run))

    # If it's a full response capture the response attributes
    if response:
        # noinspection PyBroadException
        try:
            instrumentor.set_response_attributes(span, response)
        except Exception as e:
            logger.warning(
                f"Failed to set response attributes for Baserun span, error: {e}"
            )
    else:
        # This will happen if the user doesn't return anything from their traced function
        span.set_status(
            Status(
                description="No response received",
                status_code=StatusCode.UNSET,
            )
        )

    return span



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1/baserun.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10v1/baserun.proto\x12\nbaserun.v1\x1a\x1fgoogle/protobuf/timestamp.proto\"\x9c\x01\n\x06Status\x12\x0f\n\x07message\x18\x01 \x01(\t\x12+\n\x04\x63ode\x18\x02 \x01(\x0e\x32\x1d.baserun.v1.Status.StatusCode\"T\n\nStatusCode\x12\x1b\n\x17STATUS_CODE_UNSPECIFIED\x10\x00\x12\x12\n\x0eSTATUS_CODE_OK\x10\x01\x12\x15\n\x11STATUS_CODE_ERROR\x10\x02\"V\n\x07Message\x12\x0c\n\x04role\x18\x01 \x01(\t\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\t\x12\x15\n\rfinish_reason\x18\x03 \x01(\t\x12\x15\n\rfunction_call\x18\x04 \x01(\t\"\xc7\x02\n\x03Run\x12\x0e\n\x06run_id\x18\x01 \x01(\t\x12\x10\n\x08suite_id\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0e\n\x06inputs\x18\x04 \x03(\t\x12)\n\x08run_type\x18\x05 \x01(\x0e\x32\x17.baserun.v1.Run.RunType\x12\x10\n\x08metadata\x18\x06 \x01(\t\x12\x33\n\x0fstart_timestamp\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x38\n\x14\x63ompletion_timestamp\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06result\x18\t \x01(\t\x12\r\n\x05\x65rror\x18\n \x01(\t\"5\n\x07RunType\x12\x11\n\rRUN_TYPE_TEST\x10\x00\x12\x17\n\x13RUN_TYPE_PRODUCTION\x10\x01\"c\n\x03Log\x12\x0e\n\x06run_id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07payload\x18\x03 \x01(\t\x12-\n\ttimestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xdc\x08\n\x04Span\x12\x0e\n\x06run_id\x18\x01 \x01(\t\x12\x10\n\x08trace_id\x18\x02 \x01(\x0c\x12\x0f\n\x07span_id\x18\x03 \x01(\x04\x12\x0c\n\x04name\x18\x04 \x01(\t\x12.\n\nstart_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\"\n\x06status\x18\x07 \x01(\x0b\x32\x12.baserun.v1.Status\x12\x0e\n\x06vendor\x18\x08 \x01(\t\x12\x14\n\x0crequest_type\x18\t \x01(\t\x12\r\n\x05model\x18\n \x01(\t\x12\x14\n\x0ctotal_tokens\x18\x0b \x01(\x05\x12\x19\n\x11\x63ompletion_tokens\x18\x0c \x01(\x05\x12\x15\n\rprompt_tokens\x18\r \x01(\x05\x12,\n\x0fprompt_messages\x18\x0e \x03(\x0b\x32\x13.baserun.v1.Message\x12(\n\x0b\x63ompletions\x18\x0f \x03(\x0b\x32\x13.baserun.v1.Message\x12\x15\n\x08\x61pi_base\x18\x10 \x01(\tH\x00\x88\x01\x01\x12\x15\n\x08\x61pi_type\x18\x11 \x01(\tH\x01\x88\x01\x01\x12\x16\n\tfunctions\x18\x12 \x01(\tH\x02\x88\x01\x01\x12\x1a\n\rfunction_call\x18\x13 \x01(\tH\x03\x88\x01\x01\x12\x18\n\x0btemperature\x18\x14 \x01(\x01H\x04\x88\x01\x01\x12\x12\n\x05top_p\x18\x15 \x01(\x01H\x05\x88\x01\x01\x12\x0e\n\x01n\x18\x16 \x01(\x05H\x06\x88\x01\x01\x12\x13\n\x06stream\x18\x17 \x01(\x08H\x07\x88\x01\x01\x12\x0c\n\x04stop\x18\x18 \x03(\t\x12\x17\n\nmax_tokens\x18\x19 \x01(\x05H\x08\x88\x01\x01\x12\x1d\n\x10presence_penalty\x18\x1a \x01(\x01H\t\x88\x01\x01\x12\x1e\n\x11\x66requency_penalty\x18\x1b \x01(\x01H\n\x88\x01\x01\x12\x17\n\nlogit_bias\x18\x1c \x01(\tH\x0b\x88\x01\x01\x12\x11\n\x04user\x18\x1d \x01(\tH\x0c\x88\x01\x01\x12\x15\n\x08logprobs\x18\x1e \x01(\x05H\r\x88\x01\x01\x12\x11\n\x04\x65\x63ho\x18\x1f \x01(\x08H\x0e\x88\x01\x01\x12\x13\n\x06suffix\x18  \x01(\tH\x0f\x88\x01\x01\x12\x14\n\x07\x62\x65st_of\x18! \x01(\x05H\x10\x88\x01\x01\x12\x13\n\x06log_id\x18\" \x01(\tH\x11\x88\x01\x01\x12\x12\n\x05top_k\x18# \x01(\x01H\x12\x88\x01\x01\x42\x0b\n\t_api_baseB\x0b\n\t_api_typeB\x0c\n\n_functionsB\x10\n\x0e_function_callB\x0e\n\x0c_temperatureB\x08\n\x06_top_pB\x04\n\x02_nB\t\n\x07_streamB\r\n\x0b_max_tokensB\x13\n\x11_presence_penaltyB\x14\n\x12_frequency_penaltyB\r\n\x0b_logit_biasB\x07\n\x05_userB\x0b\n\t_logprobsB\x07\n\x05_echoB\t\n\x07_suffixB\n\n\x08_best_ofB\t\n\x07_log_idB\x08\n\x06_top_k\"u\n\x04\x45val\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x0e\n\x06result\x18\x03 \x01(\t\x12\x12\n\x05score\x18\x05 \x01(\x01H\x00\x88\x01\x01\x12\x12\n\nsubmission\x18\x06 \x01(\t\x12\x0f\n\x07payload\x18\x07 \x01(\tB\x08\n\x06_score\"\x94\x01\n\tTestSuite\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x33\n\x0fstart_timestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x38\n\x14\x63ompletion_timestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"/\n\x0fStartRunRequest\x12\x1c\n\x03run\x18\x01 \x01(\x0b\x32\x0f.baserun.v1.Run\"#\n\x10StartRunResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"N\n\x10SubmitLogRequest\x12\x1c\n\x03log\x18\x01 \x01(\x0b\x32\x0f.baserun.v1.Log\x12\x1c\n\x03run\x18\x02 \x01(\x0b\x32\x0f.baserun.v1.Run\"$\n\x11SubmitLogResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"Q\n\x11SubmitSpanRequest\x12\x1e\n\x04span\x18\x01 \x01(\x0b\x32\x10.baserun.v1.Span\x12\x1c\n\x03run\x18\x02 \x01(\x0b\x32\x0f.baserun.v1.Run\"%\n\x12SubmitSpanResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"-\n\rEndRunRequest\x12\x1c\n\x03run\x18\x01 \x01(\x0b\x32\x0f.baserun.v1.Run\"!\n\x0e\x45ndRunResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"Q\n\x11SubmitEvalRequest\x12\x1e\n\x04\x65val\x18\x01 \x01(\x0b\x32\x10.baserun.v1.Eval\x12\x1c\n\x03run\x18\x02 \x01(\x0b\x32\x0f.baserun.v1.Run\"%\n\x12SubmitEvalResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"B\n\x15StartTestSuiteRequest\x12)\n\ntest_suite\x18\x01 \x01(\x0b\x32\x15.baserun.v1.TestSuite\")\n\x16StartTestSuiteResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"@\n\x13\x45ndTestSuiteRequest\x12)\n\ntest_suite\x18\x01 \x01(\x0b\x32\x15.baserun.v1.TestSuite\"\'\n\x14\x45ndTestSuiteResponse\x12\x0f\n\x07message\x18\x01 \x01(\t2\xab\x04\n\x11SubmissionService\x12\x45\n\x08StartRun\x12\x1b.baserun.v1.StartRunRequest\x1a\x1c.baserun.v1.StartRunResponse\x12H\n\tSubmitLog\x12\x1c.baserun.v1.SubmitLogRequest\x1a\x1d.baserun.v1.SubmitLogResponse\x12K\n\nSubmitSpan\x12\x1d.baserun.v1.SubmitSpanRequest\x1a\x1e.baserun.v1.SubmitSpanResponse\x12?\n\x06\x45ndRun\x12\x19.baserun.v1.EndRunRequest\x1a\x1a.baserun.v1.EndRunResponse\x12K\n\nSubmitEval\x12\x1d.baserun.v1.SubmitEvalRequest\x1a\x1e.baserun.v1.SubmitEvalResponse\x12W\n\x0eStartTestSuite\x12!.baserun.v1.StartTestSuiteRequest\x1a\".baserun.v1.StartTestSuiteResponse\x12Q\n\x0c\x45ndTestSuite\x12\x1f.baserun.v1.EndTestSuiteRequest\x1a .baserun.v1.EndTestSuiteResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'v1.baserun_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _globals['_STATUS']._serialized_start=66
  _globals['_STATUS']._serialized_end=222
  _globals['_STATUS_STATUSCODE']._serialized_start=138
  _globals['_STATUS_STATUSCODE']._serialized_end=222
  _globals['_MESSAGE']._serialized_start=224
  _globals['_MESSAGE']._serialized_end=310
  _globals['_RUN']._serialized_start=313
  _globals['_RUN']._serialized_end=640
  _globals['_RUN_RUNTYPE']._serialized_start=587
  _globals['_RUN_RUNTYPE']._serialized_end=640
  _globals['_LOG']._serialized_start=642
  _globals['_LOG']._serialized_end=741
  _globals['_SPAN']._serialized_start=744
  _globals['_SPAN']._serialized_end=1860
  _globals['_EVAL']._serialized_start=1862
  _globals['_EVAL']._serialized_end=1979
  _globals['_TESTSUITE']._serialized_start=1982
  _globals['_TESTSUITE']._serialized_end=2130
  _globals['_STARTRUNREQUEST']._serialized_start=2132
  _globals['_STARTRUNREQUEST']._serialized_end=2179
  _globals['_STARTRUNRESPONSE']._serialized_start=2181
  _globals['_STARTRUNRESPONSE']._serialized_end=2216
  _globals['_SUBMITLOGREQUEST']._serialized_start=2218
  _globals['_SUBMITLOGREQUEST']._serialized_end=2296
  _globals['_SUBMITLOGRESPONSE']._serialized_start=2298
  _globals['_SUBMITLOGRESPONSE']._serialized_end=2334
  _globals['_SUBMITSPANREQUEST']._serialized_start=2336
  _globals['_SUBMITSPANREQUEST']._serialized_end=2417
  _globals['_SUBMITSPANRESPONSE']._serialized_start=2419
  _globals['_SUBMITSPANRESPONSE']._serialized_end=2456
  _globals['_ENDRUNREQUEST']._serialized_start=2458
  _globals['_ENDRUNREQUEST']._serialized_end=2503
  _globals['_ENDRUNRESPONSE']._serialized_start=2505
  _globals['_ENDRUNRESPONSE']._serialized_end=2538
  _globals['_SUBMITEVALREQUEST']._serialized_start=2540
  _globals['_SUBMITEVALREQUEST']._serialized_end=2621
  _globals['_SUBMITEVALRESPONSE']._serialized_start=2623
  _globals['_SUBMITEVALRESPONSE']._serialized_end=2660
  _globals['_STARTTESTSUITEREQUEST']._serialized_start=2662
  _globals['_STARTTESTSUITEREQUEST']._serialized_end=2728
  _globals['_STARTTESTSUITERESPONSE']._serialized_start=2730
  _globals['_STARTTESTSUITERESPONSE']._serialized_end=2771
  _globals['_ENDTESTSUITEREQUEST']._serialized_start=2773
  _globals['_ENDTESTSUITEREQUEST']._serialized_end=2837
  _globals['_ENDTESTSUITERESPONSE']._serialized_start=2839
  _globals['_ENDTESTSUITERESPONSE']._serialized_end=2878
  _globals['_SUBMISSIONSERVICE']._serialized_start=2881
  _globals['_SUBMISSIONSERVICE']._serialized_end=3436
# @@protoc_insertion_point(module_scope)



import grpc
import os
from typing import Union
from .v1.baserun_pb2_grpc import SubmissionServiceStub
from .api_key import get_api_key


submission_service: Union[SubmissionServiceStub, None] = None


def get_or_create_submission_service():
    global submission_service
    if submission_service:
        return submission_service

    if key_chain := os.environ.get("SSL_KEY_CHAIN"):
        ssl_creds = grpc.ssl_channel_credentials(
            root_certificates=bytes(key_chain, "utf-8")
        )
    else:
        ssl_creds = grpc.ssl_channel_credentials()

    call_credentials = grpc.access_token_call_credentials(get_api_key())
    channel_credentials = grpc.composite_channel_credentials(
        ssl_creds, call_credentials
    )
    grpc_base = os.environ.get("BASERUN_GRPC_URI", "grpc.baserun.ai:50051")
    grpc_channel = grpc.secure_channel(grpc_base, channel_credentials)
    submission_service = SubmissionServiceStub(grpc_channel)
    return submission_service





#!/usr/bin/env python
import collections.abc
import logging
from typing import Collection, Any

from opentelemetry.sdk.trace import _Span

from baserun.instrumentation.base_instrumentor import BaseInstrumentor
from baserun.instrumentation.span_attributes import (
    SpanAttributes,
    ANTHROPIC_VENDOR_NAME,
)

logger = logging.getLogger(__name__)

_instruments = ("anthropic >= 0.3.0",)
__version__ = "0.1.0"


class AnthropicInstrumentor(BaseInstrumentor):
    """An instrumentor for OpenAI's client library."""

    @staticmethod
    def wrapped_methods() -> list[dict[str, Any]]:
        from anthropic.resources import Completions, AsyncCompletions

        return [
            {
                "class": Completions,
                "function": Completions.create,
                "span_name": "anthropic.completion",
            },
            {
                "class": AsyncCompletions,
                "function": AsyncCompletions.create,
                "span_name": "anthropic.completion",
            },
        ]

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    @staticmethod
    def set_request_attributes(span: _Span, kwargs: dict[str, Any]):
        span.set_attribute(SpanAttributes.LLM_VENDOR, ANTHROPIC_VENDOR_NAME)
        span.set_attribute(SpanAttributes.LLM_REQUEST_MODEL, kwargs.get("model"))

        if "top_k" in kwargs:
            span.set_attribute(SpanAttributes.LLM_TOP_K, kwargs.get("top_k"))

        if "top_p" in kwargs:
            span.set_attribute(SpanAttributes.LLM_TOP_P, kwargs.get("top_p"))

        if "stream" in kwargs:
            span.set_attribute(SpanAttributes.LLM_STREAM, kwargs.get("stream"))

        if "temperature" in kwargs:
            span.set_attribute(
                SpanAttributes.LLM_TEMPERATURE, kwargs.get("temperature")
            )

        if stop_sequences := kwargs.get("stop_sequences"):
            span.set_attribute(SpanAttributes.LLM_CHAT_STOP_SEQUENCES, stop_sequences)

        span.set_attribute(
            SpanAttributes.LLM_REQUEST_MAX_TOKENS, kwargs.get("max_tokens_to_sample")
        )

        prefix = f"{SpanAttributes.LLM_PROMPTS}.0"
        span.set_attribute(f"{prefix}.content", kwargs.get("prompt"))

    @staticmethod
    def set_response_attributes(span: _Span, response):
        prefix = f"{SpanAttributes.LLM_COMPLETIONS}.0"
        span.set_attribute(f"{prefix}.finish_reason", response.stop_reason)
        span.set_attribute(f"{prefix}.content", response.completion)
        span.set_attribute(SpanAttributes.ANTHROPIC_LOG_ID, response.log_id)

    @staticmethod
    def generator_wrapper(original_generator: collections.abc.Iterator, span: _Span):
        for value in original_generator:
            AnthropicInstrumentor._handle_generator_value(value, span)

            yield value

    @staticmethod
    async def async_generator_wrapper(
        original_generator: collections.abc.AsyncIterator, span: _Span
    ):
        async for value in original_generator:
            AnthropicInstrumentor._handle_generator_value(value, span)

            yield value

    @staticmethod
    def _handle_generator_value(value, span: _Span):
        new_content = value.completion

        prefix = f"{SpanAttributes.LLM_COMPLETIONS}.0"
        content_attribute = f"{prefix}.content"

        if new_content:
            content = span.attributes.get(content_attribute, "")
            span.set_attribute(content_attribute, content + new_content)

        if new_content is None or value.stop_reason:
            span.end()



import os
from unittest.mock import patch, call

import openai
import pytest
from dotenv import load_dotenv
from opentelemetry.sdk.trace.export import SimpleSpanProcessor

from baserun import Baserun
from baserun.v1.baserun_pb2 import Run, Span, StartRunRequest

load_dotenv()


@pytest.fixture(autouse=True)
def set_openai_api_key():
    api_key = os.environ.get("OPENAI_API_KEY")

    openai.api_key = api_key


@pytest.fixture
def mock_services():
    with patch("baserun.grpc.submission_service.StartRun") as mock_start_run, patch(
        "baserun.grpc.submission_service.SubmitLog"
    ) as mock_submit_log, patch(
        "baserun.grpc.submission_service.SubmitSpan"
    ) as mock_submit_span, patch(
        "baserun.grpc.submission_service.EndRun"
    ) as mock_end_run, patch(
        "baserun.grpc.submission_service.SubmitEval"
    ) as mock_submit_eval, patch(
        "baserun.grpc.submission_service.StartTestSuite"
    ) as mock_start_test_suite, patch(
        "baserun.grpc.submission_service.EndTestSuite"
    ) as mock_end_test_suite:
        yield {
            "mock_start_run": mock_start_run,
            "mock_submit_log": mock_submit_log,
            "mock_submit_span": mock_submit_span,
            "mock_end_run": mock_end_run,
            "mock_submit_eval": mock_submit_eval,
            "mock_start_test_suite": mock_start_test_suite,
            "mock_end_test_suite": mock_end_test_suite,
        }


def pytest_sessionstart(session):
    """Starting up Baserun in tests requires that these things happen in a specific order:
    - `init`, specifically setting up gRPC
    - Mock services, to replace the services that were just set up
    - Instrument
    - Close channel, simply to ensure that no unmocked calls get through
    """
    Baserun.init(instrument=False)
    # mock_services()
    # Replace the batch processor so that things happen synchronously and not in a separate thread
    Baserun.instrument(processor_class=SimpleSpanProcessor)


def get_mock_objects(mock_services) -> tuple[Run, Span, Run, Run]:
    mock_services["mock_start_run"].assert_called_once()
    run_call: call = mock_services["mock_start_run"].call_args_list[0]
    start_run_request: StartRunRequest = run_call.args[0]
    started_run = start_run_request.run

    if len(mock_services["mock_submit_span"].call_args_list):
        mock_services["mock_submit_span"].assert_called_once()
        submit_span_call: call = mock_services["mock_submit_span"].call_args_list[0]
        submit_span_request = submit_span_call.args[0]
        span = submit_span_request.span
        submitted_run = submit_span_request.run
    else:
        span = None
        submitted_run = None

    mock_services["mock_end_run"].assert_called_once()
    end_run_call: call = mock_services["mock_end_run"].call_args_list[0]
    end_run_request: StartRunRequest = end_run_call.args[0]
    ended_run = end_run_request.run

    return started_run, span, submitted_run, ended_run



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from baserun.v1 import baserun_pb2 as v1_dot_baserun__pb2


class SubmissionServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.StartRun = channel.unary_unary(
                '/baserun.v1.SubmissionService/StartRun',
                request_serializer=v1_dot_baserun__pb2.StartRunRequest.SerializeToString,
                response_deserializer=v1_dot_baserun__pb2.StartRunResponse.FromString,
                )
        self.SubmitLog = channel.unary_unary(
                '/baserun.v1.SubmissionService/SubmitLog',
                request_serializer=v1_dot_baserun__pb2.SubmitLogRequest.SerializeToString,
                response_deserializer=v1_dot_baserun__pb2.SubmitLogResponse.FromString,
                )
        self.SubmitSpan = channel.unary_unary(
                '/baserun.v1.SubmissionService/SubmitSpan',
                request_serializer=v1_dot_baserun__pb2.SubmitSpanRequest.SerializeToString,
                response_deserializer=v1_dot_baserun__pb2.SubmitSpanResponse.FromString,
                )
        self.EndRun = channel.unary_unary(
                '/baserun.v1.SubmissionService/EndRun',
                request_serializer=v1_dot_baserun__pb2.EndRunRequest.SerializeToString,
                response_deserializer=v1_dot_baserun__pb2.EndRunResponse.FromString,
                )
        self.SubmitEval = channel.unary_unary(
                '/baserun.v1.SubmissionService/SubmitEval',
                request_serializer=v1_dot_baserun__pb2.SubmitEvalRequest.SerializeToString,
                response_deserializer=v1_dot_baserun__pb2.SubmitEvalResponse.FromString,
                )
        self.StartTestSuite = channel.unary_unary(
                '/baserun.v1.SubmissionService/StartTestSuite',
                request_serializer=v1_dot_baserun__pb2.StartTestSuiteRequest.SerializeToString,
                response_deserializer=v1_dot_baserun__pb2.StartTestSuiteResponse.FromString,
                )
        self.EndTestSuite = channel.unary_unary(
                '/baserun.v1.SubmissionService/EndTestSuite',
                request_serializer=v1_dot_baserun__pb2.EndTestSuiteRequest.SerializeToString,
                response_deserializer=v1_dot_baserun__pb2.EndTestSuiteResponse.FromString,
                )


class SubmissionServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def StartRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubmitLog(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubmitSpan(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EndRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubmitEval(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StartTestSuite(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EndTestSuite(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SubmissionServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'StartRun': grpc.unary_unary_rpc_method_handler(
                    servicer.StartRun,
                    request_deserializer=v1_dot_baserun__pb2.StartRunRequest.FromString,
                    response_serializer=v1_dot_baserun__pb2.StartRunResponse.SerializeToString,
            ),
            'SubmitLog': grpc.unary_unary_rpc_method_handler(
                    servicer.SubmitLog,
                    request_deserializer=v1_dot_baserun__pb2.SubmitLogRequest.FromString,
                    response_serializer=v1_dot_baserun__pb2.SubmitLogResponse.SerializeToString,
            ),
            'SubmitSpan': grpc.unary_unary_rpc_method_handler(
                    servicer.SubmitSpan,
                    request_deserializer=v1_dot_baserun__pb2.SubmitSpanRequest.FromString,
                    response_serializer=v1_dot_baserun__pb2.SubmitSpanResponse.SerializeToString,
            ),
            'EndRun': grpc.unary_unary_rpc_method_handler(
                    servicer.EndRun,
                    request_deserializer=v1_dot_baserun__pb2.EndRunRequest.FromString,
                    response_serializer=v1_dot_baserun__pb2.EndRunResponse.SerializeToString,
            ),
            'SubmitEval': grpc.unary_unary_rpc_method_handler(
                    servicer.SubmitEval,
                    request_deserializer=v1_dot_baserun__pb2.SubmitEvalRequest.FromString,
                    response_serializer=v1_dot_baserun__pb2.SubmitEvalResponse.SerializeToString,
            ),
            'StartTestSuite': grpc.unary_unary_rpc_method_handler(
                    servicer.StartTestSuite,
                    request_deserializer=v1_dot_baserun__pb2.StartTestSuiteRequest.FromString,
                    response_serializer=v1_dot_baserun__pb2.StartTestSuiteResponse.SerializeToString,
            ),
            'EndTestSuite': grpc.unary_unary_rpc_method_handler(
                    servicer.EndTestSuite,
                    request_deserializer=v1_dot_baserun__pb2.EndTestSuiteRequest.FromString,
                    response_serializer=v1_dot_baserun__pb2.EndTestSuiteResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'baserun.v1.SubmissionService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class SubmissionService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def StartRun(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/baserun.v1.SubmissionService/StartRun',
            v1_dot_baserun__pb2.StartRunRequest.SerializeToString,
            v1_dot_baserun__pb2.StartRunResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SubmitLog(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/baserun.v1.SubmissionService/SubmitLog',
            v1_dot_baserun__pb2.SubmitLogRequest.SerializeToString,
            v1_dot_baserun__pb2.SubmitLogResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SubmitSpan(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/baserun.v1.SubmissionService/SubmitSpan',
            v1_dot_baserun__pb2.SubmitSpanRequest.SerializeToString,
            v1_dot_baserun__pb2.SubmitSpanResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def EndRun(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/baserun.v1.SubmissionService/EndRun',
            v1_dot_baserun__pb2.EndRunRequest.SerializeToString,
            v1_dot_baserun__pb2.EndRunResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SubmitEval(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/baserun.v1.SubmissionService/SubmitEval',
            v1_dot_baserun__pb2.SubmitEvalRequest.SerializeToString,
            v1_dot_baserun__pb2.SubmitEvalResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StartTestSuite(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/baserun.v1.SubmissionService/StartTestSuite',
            v1_dot_baserun__pb2.StartTestSuiteRequest.SerializeToString,
            v1_dot_baserun__pb2.StartTestSuiteResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def EndTestSuite(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/baserun.v1.SubmissionService/EndTestSuite',
            v1_dot_baserun__pb2.EndTestSuiteRequest.SerializeToString,
            v1_dot_baserun__pb2.EndTestSuiteResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import argparse
import asyncio
import inspect
import os
from threading import Thread

import openai
from openai import ChatCompletion, Completion

import baserun


@baserun.trace
def openai_chat(prompt="What is the capitol of the US?") -> str:
    completion = ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[{"role": "user", "content": prompt}],
    )
    return completion.choices[0]["message"].content


def openai_chat_unwrapped(prompt="What is the capitol of the US?", **kwargs) -> str:
    completion = ChatCompletion.create(
        model="gpt-3.5-turbo", messages=[{"role": "user", "content": prompt}], **kwargs
    )
    result = completion.choices[0]["message"].content
    print(result)
    return result


@baserun.trace
async def openai_chat_async(prompt="What is the capitol of the US?") -> str:
    completion = await ChatCompletion.acreate(
        model="gpt-3.5-turbo",
        messages=[{"role": "user", "content": prompt}],
    )
    return completion.choices[0]["message"].content


@baserun.trace
def openai_chat_functions(prompt="Say 'hello world'") -> dict[str, str]:
    functions = [
        {
            "name": "say",
            "description": "Convert some text to speech",
            "parameters": {
                "type": "object",
                "properties": {
                    "text": {"type": "string", "description": "The text to speak"},
                },
                "required": ["text"],
            },
        }
    ]
    completion = ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[{"role": "user", "content": prompt}],
        functions=functions,
        function_call={"name": "say"},
    )
    return completion.choices[0]["message"].function_call


@baserun.trace
def openai_chat_functions_streaming(prompt="Say 'hello world'") -> dict[str, str]:
    functions = [
        {
            "name": "say",
            "description": "Convert some text to speech",
            "parameters": {
                "type": "object",
                "properties": {
                    "text": {"type": "string", "description": "The text to speak"},
                },
                "required": ["text"],
            },
        }
    ]
    completion_generator = ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[{"role": "user", "content": prompt}],
        functions=functions,
        stream=True,
        function_call={"name": "say"},
    )
    function_name = ""
    function_arguments = ""
    for chunk in completion_generator:
        choice = chunk.choices[0]
        if function_call := choice.delta.get("function_call"):
            function_name += function_call.get("name", "")
            function_arguments += function_call.get("arguments", "")

    return {"name": function_name, "arguments": function_arguments}


@baserun.trace
def openai_chat_streaming(prompt="What is the capitol of the US?") -> str:
    completion_generator = ChatCompletion.create(
        model="gpt-3.5-turbo",
        stream=True,
        messages=[{"role": "user", "content": prompt}],
    )
    content = ""
    for chunk in completion_generator:
        if new_content := chunk.choices[0].delta.get("content"):
            content += new_content

    return content


@baserun.trace
async def openai_chat_async_streaming(prompt="What is the capitol of the US?") -> str:
    completion_generator = await ChatCompletion.acreate(
        model="gpt-3.5-turbo",
        stream=True,
        messages=[{"role": "user", "content": prompt}],
    )
    content = ""
    async for chunk in completion_generator:
        if new_content := chunk.choices[0].delta.get("content"):
            content += new_content

    return content


@baserun.trace
def openai_chat_error(prompt="What is the capitol of the US?"):
    original_api_type = openai.api_type
    try:
        openai.api_type = "somegarbage"
        # Will raise InvalidAPIType
        ChatCompletion.create(
            model="gpt-3.5-turbo",
            messages=[{"role": "user", "content": prompt}],
        )
    finally:
        openai.api_type = original_api_type


@baserun.trace
def traced_fn_error():
    ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[{"role": "user", "content": "What is the capitol of the US?"}],
    )
    raise ValueError("Something went wrong")


@baserun.trace
def openai_completion(prompt="Human: say this is a test\nAssistant: ") -> str:
    completion = Completion.create(model="text-davinci-003", prompt=prompt)
    return completion.choices[0].text


@baserun.trace
async def openai_completion_async(
    prompt="Human: say this is a test\nAssistant: ",
) -> str:
    completion = await Completion.acreate(model="text-davinci-003", prompt=prompt)
    return completion.choices[0].text


@baserun.trace
def openai_completion_streaming(prompt="Human: say this is a test\nAssistant: ") -> str:
    completion_generator = Completion.create(
        model="text-davinci-003", prompt=prompt, stream=True
    )

    content = ""
    for chunk in completion_generator:
        if new_content := chunk.choices[0].text:
            content += new_content

    return content


@baserun.trace
async def openai_completion_async_streaming(
    prompt="Human: say this is a test\nAssistant: ",
) -> str:
    completion_generator = await Completion.acreate(
        model="text-davinci-003", prompt=prompt, stream=True
    )
    content = ""
    async for chunk in completion_generator:
        if new_content := chunk.choices[0].text:
            content += new_content

    return content


@baserun.trace
def openai_threaded():
    threads = [
        Thread(
            target=baserun.thread_wrapper(openai_chat_unwrapped),
            args=("What is the capitol of the state of Georgia?",),
        ),
        Thread(
            target=baserun.thread_wrapper(openai_chat_unwrapped),
            args=("What is the capitol of the California?",),
            kwargs={"top_p": 0.5},
        ),
        Thread(
            target=baserun.thread_wrapper(openai_chat_unwrapped),
            args=("What is the capitol of the Montana?",),
            kwargs={"temperature": 1},
        ),
    ]
    [t.start() for t in threads]
    [t.join() for t in threads]


# Allows you to call any of these functions, e.g. python tests/testing_functions.py openai_chat_functions_streaming
if __name__ == "__main__":
    from dotenv import load_dotenv
    from baserun import Baserun

    load_dotenv()
    openai.api_key = os.environ.get("OPENAI_API_KEY")
    Baserun.init()

    parser = argparse.ArgumentParser(description="Execute a function with a prompt.")
    parser.add_argument(
        "function_to_call", type=str, help="Name of the function to call"
    )
    parser.add_argument(
        "--prompt", type=str, help="Prompt to pass to the function", default=None
    )

    parsed_args = parser.parse_args()

    # Resolve the string function name to the function object
    function_to_call = globals().get(parsed_args.function_to_call)
    if inspect.iscoroutinefunction(function_to_call):
        if parsed_args.prompt:
            result = asyncio.run(function_to_call(parsed_args.prompt))
        else:
            result = asyncio.run(function_to_call())
    else:
        if parsed_args.prompt:
            result = function_to_call(parsed_args.prompt)
        else:
            result = function_to_call()

    print(result)



import os

from .baserun import Baserun
from .thread_wrapper import baserun_thread_wrapper

init = Baserun.init
trace = Baserun.trace
log = Baserun.log
evals = Baserun.evals
thread_wrapper = baserun_thread_wrapper
api_key = os.environ.get("BASERUN_API_KEY")



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Status(_message.Message):
    __slots__ = ["message", "code"]
    class StatusCode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        STATUS_CODE_UNSPECIFIED: _ClassVar[Status.StatusCode]
        STATUS_CODE_OK: _ClassVar[Status.StatusCode]
        STATUS_CODE_ERROR: _ClassVar[Status.StatusCode]
    STATUS_CODE_UNSPECIFIED: Status.StatusCode
    STATUS_CODE_OK: Status.StatusCode
    STATUS_CODE_ERROR: Status.StatusCode
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    CODE_FIELD_NUMBER: _ClassVar[int]
    message: str
    code: Status.StatusCode
    def __init__(self, message: _Optional[str] = ..., code: _Optional[_Union[Status.StatusCode, str]] = ...) -> None: ...

class Message(_message.Message):
    __slots__ = ["role", "content", "finish_reason", "function_call"]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    FINISH_REASON_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_CALL_FIELD_NUMBER: _ClassVar[int]
    role: str
    content: str
    finish_reason: str
    function_call: str
    def __init__(self, role: _Optional[str] = ..., content: _Optional[str] = ..., finish_reason: _Optional[str] = ..., function_call: _Optional[str] = ...) -> None: ...

class Run(_message.Message):
    __slots__ = ["run_id", "suite_id", "name", "inputs", "run_type", "metadata", "start_timestamp", "completion_timestamp", "result", "error"]
    class RunType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        RUN_TYPE_TEST: _ClassVar[Run.RunType]
        RUN_TYPE_PRODUCTION: _ClassVar[Run.RunType]
    RUN_TYPE_TEST: Run.RunType
    RUN_TYPE_PRODUCTION: Run.RunType
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    SUITE_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    RUN_TYPE_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    START_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    COMPLETION_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    run_id: str
    suite_id: str
    name: str
    inputs: _containers.RepeatedScalarFieldContainer[str]
    run_type: Run.RunType
    metadata: str
    start_timestamp: _timestamp_pb2.Timestamp
    completion_timestamp: _timestamp_pb2.Timestamp
    result: str
    error: str
    def __init__(self, run_id: _Optional[str] = ..., suite_id: _Optional[str] = ..., name: _Optional[str] = ..., inputs: _Optional[_Iterable[str]] = ..., run_type: _Optional[_Union[Run.RunType, str]] = ..., metadata: _Optional[str] = ..., start_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., completion_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., result: _Optional[str] = ..., error: _Optional[str] = ...) -> None: ...

class Log(_message.Message):
    __slots__ = ["run_id", "name", "payload", "timestamp"]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    run_id: str
    name: str
    payload: str
    timestamp: _timestamp_pb2.Timestamp
    def __init__(self, run_id: _Optional[str] = ..., name: _Optional[str] = ..., payload: _Optional[str] = ..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class Span(_message.Message):
    __slots__ = ["run_id", "trace_id", "span_id", "name", "start_time", "end_time", "status", "vendor", "request_type", "model", "total_tokens", "completion_tokens", "prompt_tokens", "prompt_messages", "completions", "api_base", "api_type", "functions", "function_call", "temperature", "top_p", "n", "stream", "stop", "max_tokens", "presence_penalty", "frequency_penalty", "logit_bias", "user", "logprobs", "echo", "suffix", "best_of", "log_id", "top_k"]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    SPAN_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    VENDOR_FIELD_NUMBER: _ClassVar[int]
    REQUEST_TYPE_FIELD_NUMBER: _ClassVar[int]
    MODEL_FIELD_NUMBER: _ClassVar[int]
    TOTAL_TOKENS_FIELD_NUMBER: _ClassVar[int]
    COMPLETION_TOKENS_FIELD_NUMBER: _ClassVar[int]
    PROMPT_TOKENS_FIELD_NUMBER: _ClassVar[int]
    PROMPT_MESSAGES_FIELD_NUMBER: _ClassVar[int]
    COMPLETIONS_FIELD_NUMBER: _ClassVar[int]
    API_BASE_FIELD_NUMBER: _ClassVar[int]
    API_TYPE_FIELD_NUMBER: _ClassVar[int]
    FUNCTIONS_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_CALL_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_FIELD_NUMBER: _ClassVar[int]
    TOP_P_FIELD_NUMBER: _ClassVar[int]
    N_FIELD_NUMBER: _ClassVar[int]
    STREAM_FIELD_NUMBER: _ClassVar[int]
    STOP_FIELD_NUMBER: _ClassVar[int]
    MAX_TOKENS_FIELD_NUMBER: _ClassVar[int]
    PRESENCE_PENALTY_FIELD_NUMBER: _ClassVar[int]
    FREQUENCY_PENALTY_FIELD_NUMBER: _ClassVar[int]
    LOGIT_BIAS_FIELD_NUMBER: _ClassVar[int]
    USER_FIELD_NUMBER: _ClassVar[int]
    LOGPROBS_FIELD_NUMBER: _ClassVar[int]
    ECHO_FIELD_NUMBER: _ClassVar[int]
    SUFFIX_FIELD_NUMBER: _ClassVar[int]
    BEST_OF_FIELD_NUMBER: _ClassVar[int]
    LOG_ID_FIELD_NUMBER: _ClassVar[int]
    TOP_K_FIELD_NUMBER: _ClassVar[int]
    run_id: str
    trace_id: bytes
    span_id: int
    name: str
    start_time: _timestamp_pb2.Timestamp
    end_time: _timestamp_pb2.Timestamp
    status: Status
    vendor: str
    request_type: str
    model: str
    total_tokens: int
    completion_tokens: int
    prompt_tokens: int
    prompt_messages: _containers.RepeatedCompositeFieldContainer[Message]
    completions: _containers.RepeatedCompositeFieldContainer[Message]
    api_base: str
    api_type: str
    functions: str
    function_call: str
    temperature: float
    top_p: float
    n: int
    stream: bool
    stop: _containers.RepeatedScalarFieldContainer[str]
    max_tokens: int
    presence_penalty: float
    frequency_penalty: float
    logit_bias: str
    user: str
    logprobs: int
    echo: bool
    suffix: str
    best_of: int
    log_id: str
    top_k: float
    def __init__(self, run_id: _Optional[str] = ..., trace_id: _Optional[bytes] = ..., span_id: _Optional[int] = ..., name: _Optional[str] = ..., start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., status: _Optional[_Union[Status, _Mapping]] = ..., vendor: _Optional[str] = ..., request_type: _Optional[str] = ..., model: _Optional[str] = ..., total_tokens: _Optional[int] = ..., completion_tokens: _Optional[int] = ..., prompt_tokens: _Optional[int] = ..., prompt_messages: _Optional[_Iterable[_Union[Message, _Mapping]]] = ..., completions: _Optional[_Iterable[_Union[Message, _Mapping]]] = ..., api_base: _Optional[str] = ..., api_type: _Optional[str] = ..., functions: _Optional[str] = ..., function_call: _Optional[str] = ..., temperature: _Optional[float] = ..., top_p: _Optional[float] = ..., n: _Optional[int] = ..., stream: bool = ..., stop: _Optional[_Iterable[str]] = ..., max_tokens: _Optional[int] = ..., presence_penalty: _Optional[float] = ..., frequency_penalty: _Optional[float] = ..., logit_bias: _Optional[str] = ..., user: _Optional[str] = ..., logprobs: _Optional[int] = ..., echo: bool = ..., suffix: _Optional[str] = ..., best_of: _Optional[int] = ..., log_id: _Optional[str] = ..., top_k: _Optional[float] = ...) -> None: ...

class Eval(_message.Message):
    __slots__ = ["name", "type", "result", "score", "submission", "payload"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    SCORE_FIELD_NUMBER: _ClassVar[int]
    SUBMISSION_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    name: str
    type: str
    result: str
    score: float
    submission: str
    payload: str
    def __init__(self, name: _Optional[str] = ..., type: _Optional[str] = ..., result: _Optional[str] = ..., score: _Optional[float] = ..., submission: _Optional[str] = ..., payload: _Optional[str] = ...) -> None: ...

class TestSuite(_message.Message):
    __slots__ = ["id", "name", "start_timestamp", "completion_timestamp"]
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    START_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    COMPLETION_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    start_timestamp: _timestamp_pb2.Timestamp
    completion_timestamp: _timestamp_pb2.Timestamp
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., start_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., completion_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class StartRunRequest(_message.Message):
    __slots__ = ["run"]
    RUN_FIELD_NUMBER: _ClassVar[int]
    run: Run
    def __init__(self, run: _Optional[_Union[Run, _Mapping]] = ...) -> None: ...

class StartRunResponse(_message.Message):
    __slots__ = ["message"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class SubmitLogRequest(_message.Message):
    __slots__ = ["log", "run"]
    LOG_FIELD_NUMBER: _ClassVar[int]
    RUN_FIELD_NUMBER: _ClassVar[int]
    log: Log
    run: Run
    def __init__(self, log: _Optional[_Union[Log, _Mapping]] = ..., run: _Optional[_Union[Run, _Mapping]] = ...) -> None: ...

class SubmitLogResponse(_message.Message):
    __slots__ = ["message"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class SubmitSpanRequest(_message.Message):
    __slots__ = ["span", "run"]
    SPAN_FIELD_NUMBER: _ClassVar[int]
    RUN_FIELD_NUMBER: _ClassVar[int]
    span: Span
    run: Run
    def __init__(self, span: _Optional[_Union[Span, _Mapping]] = ..., run: _Optional[_Union[Run, _Mapping]] = ...) -> None: ...

class SubmitSpanResponse(_message.Message):
    __slots__ = ["message"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class EndRunRequest(_message.Message):
    __slots__ = ["run"]
    RUN_FIELD_NUMBER: _ClassVar[int]
    run: Run
    def __init__(self, run: _Optional[_Union[Run, _Mapping]] = ...) -> None: ...

class EndRunResponse(_message.Message):
    __slots__ = ["message"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class SubmitEvalRequest(_message.Message):
    __slots__ = ["eval", "run"]
    EVAL_FIELD_NUMBER: _ClassVar[int]
    RUN_FIELD_NUMBER: _ClassVar[int]
    eval: Eval
    run: Run
    def __init__(self, eval: _Optional[_Union[Eval, _Mapping]] = ..., run: _Optional[_Union[Run, _Mapping]] = ...) -> None: ...

class SubmitEvalResponse(_message.Message):
    __slots__ = ["message"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class StartTestSuiteRequest(_message.Message):
    __slots__ = ["test_suite"]
    TEST_SUITE_FIELD_NUMBER: _ClassVar[int]
    test_suite: TestSuite
    def __init__(self, test_suite: _Optional[_Union[TestSuite, _Mapping]] = ...) -> None: ...

class StartTestSuiteResponse(_message.Message):
    __slots__ = ["message"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class EndTestSuiteRequest(_message.Message):
    __slots__ = ["test_suite"]
    TEST_SUITE_FIELD_NUMBER: _ClassVar[int]
    test_suite: TestSuite
    def __init__(self, test_suite: _Optional[_Union[TestSuite, _Mapping]] = ...) -> None: ...

class EndTestSuiteResponse(_message.Message):
    __slots__ = ["message"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...



import pytest
from google import protobuf
from openai.error import InvalidAPIType
from opentelemetry.trace import StatusCode

from baserun.v1.baserun_pb2 import Run, Span
from tests.conftest import get_mock_objects
from tests.testing_functions import (
    openai_chat,
    openai_chat_async,
    openai_chat_functions,
    openai_chat_functions_streaming,
    openai_chat_streaming,
    openai_chat_error,
    traced_fn_error,
    openai_completion,
    openai_chat_async_streaming,
    openai_completion_async,
)


def basic_run_asserts(run: Run, name: str = "", result: str = "", error: str = ""):
    assert run.name == name
    assert result in run.result
    assert error in run.error

    assert isinstance(run.run_id, str)
    assert isinstance(run.suite_id, str)
    assert run.inputs == []
    assert run.run_type == Run.RunType.RUN_TYPE_TEST
    assert run.metadata == "{}"
    assert isinstance(run.completion_timestamp, protobuf.timestamp_pb2.Timestamp)
    assert isinstance(run.start_timestamp, protobuf.timestamp_pb2.Timestamp)


def basic_span_asserts(
    span: Span,
    request_type="chat",
    status_code=StatusCode.OK.value,
    api_type="open_ai",
    prompt_role="user",
    completion_role="assistant",
    prompt="What is the capitol of the US?",
    result: str = "washington",
):
    assert isinstance(span.run_id, str)
    assert isinstance(span.trace_id, bytes)
    assert isinstance(span.span_id, int)
    assert span.name == f"baserun.openai.{request_type}"
    assert isinstance(span.start_time, protobuf.timestamp_pb2.Timestamp)
    assert isinstance(span.end_time, protobuf.timestamp_pb2.Timestamp)
    assert span.status.code == status_code
    assert span.vendor == "OpenAI"
    assert span.request_type == request_type
    assert span.api_base == "https://api.openai.com/v1"
    assert span.api_type == api_type
    assert isinstance(span.total_tokens, int)
    assert isinstance(span.completion_tokens, int)
    assert isinstance(span.prompt_tokens, int)
    assert span.stop == []
    assert span.max_tokens == 0

    prompt_message = span.prompt_messages[0]
    assert prompt_message.role == prompt_role
    assert prompt_message.content == prompt

    # Some things, like error responses, don't have any completions
    if len(span.completions):
        completion = span.completions[0]
        assert completion.role == completion_role
        assert result.lower() in completion.content.lower()


def test_chat_completion_basic(mock_services):
    name = "test_chat_completion_basic"
    openai_chat()

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="Washington")

    basic_span_asserts(span)


@pytest.mark.asyncio
async def test_chat_completion_async(mock_services):
    name = "test_chat_completion_async"
    await openai_chat_async()

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="Washington")

    basic_span_asserts(span)


@pytest.mark.asyncio
async def test_chat_completion_async_streaming(mock_services):
    name = "test_chat_completion_async_streaming"
    await openai_chat_async_streaming()

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="Washington")

    basic_span_asserts(span)


def test_openai_chat_with_functions_streaming(mock_services):
    name = "test_openai_chat_with_functions_streaming"
    prompt = "Say 'hello world'"
    function_call = openai_chat_functions_streaming(prompt=prompt)
    assert function_call.get("name") == "say"
    assert "hello world" in function_call.get("arguments")

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="")

    basic_span_asserts(span, prompt=prompt, result="")

    # Request parameters
    assert span.function_call == '{"name": "say"}'
    assert "text to speech" in span.functions

    # Response parameters (which is on the completion and not on the span itself)
    completion = span.completions[0]
    assert completion.role == "assistant"
    assert "say" in completion.function_call
    assert "hello world" in completion.function_call


def test_openai_chat_with_functions(mock_services):
    name = "test_openai_chat_with_functions"
    prompt = "Say 'hello world'"
    function_call = openai_chat_functions(prompt=prompt)
    assert function_call.get("name") == "say"
    assert "hello world" in function_call.get("arguments")

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="")

    basic_span_asserts(span, prompt=prompt, result="")

    # Request parameters
    assert span.function_call == '{"name": "say"}'
    assert "text to speech" in span.functions

    # Response parameters (which is on the completion and not on the span itself)
    completion = span.completions[0]
    assert completion.role == "assistant"
    assert "say" in completion.function_call
    assert "hello world" in completion.function_call


def test_chat_completion_streaming(mock_services):
    name = "test_chat_completion_streaming"
    openai_chat_streaming()
    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="Washington")

    basic_span_asserts(span)


def test_chat_completion_error(mock_services):
    """Tests when a call to OpenAI fails (rate limit, service unavailable, etc)"""
    name = "test_chat_completion_error"
    with pytest.raises(InvalidAPIType):
        openai_chat_error()

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="", error="InvalidAPIType")

    basic_span_asserts(span, status_code=StatusCode.ERROR.value, api_type="somegarbage")


def test_traced_fn_error(mock_services):
    """Tests when calls to OpenAI succeed, but an error is raised in the user's application logic"""
    name = "test_traced_fn_error"
    with pytest.raises(ValueError):
        traced_fn_error()

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="", error="Something went wrong")

    # The run itself failed but the OpenAI call was successful (see `test_chat_completion_error` for a failed span)
    basic_span_asserts(span, status_code=StatusCode.OK.value)


def test_completion(mock_services):
    prompt = "Human: say this is a test\nAssistant: "
    name = "test_completion"
    openai_completion(prompt)
    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="test")

    basic_span_asserts(
        span,
        request_type="completion",
        prompt=prompt,
        prompt_role="",
        completion_role="",
        result="test",
    )


@pytest.mark.asyncio
async def test_completion_async(mock_services):
    prompt = "Human: say this is a test\nAssistant: "
    name = "test_completion_async"
    await openai_completion_async(prompt)
    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="test")

    basic_span_asserts(
        span,
        request_type="completion",
        prompt=prompt,
        prompt_role="",
        completion_role="",
        result="test",
    )



class SpanAttributes:
    BASERUN_RUN = "baserun.run"
    # Matches opentelemetry-semconv-llm
    LLM_VENDOR = "llm.vendor"
    LLM_REQUEST_TYPE = "llm.request.type"
    LLM_REQUEST_MODEL = "llm.request.model"
    LLM_RESPONSE_MODEL = "llm.response.model"
    LLM_REQUEST_MAX_TOKENS = "llm.request.max_tokens"
    LLM_USAGE_TOTAL_TOKENS = "llm.usage.total_tokens"
    LLM_USAGE_COMPLETION_TOKENS = "llm.usage.completion_tokens"
    LLM_USAGE_PROMPT_TOKENS = "llm.usage.prompt_tokens"
    LLM_TEMPERATURE = "llm.temperature"
    LLM_TOP_P = "llm.top_p"
    LLM_TOP_K = "llm.top_k"
    LLM_FREQUENCY_PENALTY = "llm.frequency_penalty"
    LLM_PRESENCE_PENALTY = "llm.presence_penalty"
    LLM_PROMPTS = "llm.prompts"
    LLM_COMPLETIONS = "llm.completions"
    LLM_CHAT_STOP_SEQUENCES = "llm.chat.stop_sequences"
    LLM_FUNCTION_CALL = "llm.function_call"
    LLM_FUNCTIONS = "llm.functions"
    LLM_N = "llm.n"
    LLM_STREAM = "llm.stream"
    LLM_STOP = "llm.stop"
    LLM_LOGIT_BIAS = "llm.logit_bias"
    LLM_USER = "llm.user"
    LLM_BEST_OF = "llm.best_of"
    LLM_LOGPROBS = "llm.logprobs"
    LLM_SUFFIX = "llm.suffix"
    LLM_ECHO = "llm.echo"

    OPENAI_API_VERSION = "openai.api_version"
    OPENAI_API_BASE = "openai.api_base"
    OPENAI_API_TYPE = "openai.api_type"

    ANTHROPIC_LOG_ID = "anthropic.log_id"


ANTHROPIC_VENDOR_NAME = "Anthropic"
OPENAI_VENDOR_NAME = "OpenAI"



import collections.abc
import inspect
from abc import abstractmethod
from typing import Any, Callable

from opentelemetry.instrumentation.instrumentor import (
    BaseInstrumentor as OTelInstrumentor,
)
from opentelemetry.sdk.trace import Span, _Span

from baserun.instrumentation.instrumented_wrapper import (
    instrumented_wrapper,
    async_instrumented_wrapper,
)


class BaseInstrumentor(OTelInstrumentor):
    original_methods: dict[str, Callable] = None

    @staticmethod
    @abstractmethod
    def wrapped_methods() -> list[dict[str, Any]]:
        pass

    @staticmethod
    @abstractmethod
    def set_request_attributes(span: Span, kwargs: dict[str, Any]):
        pass

    @staticmethod
    @abstractmethod
    def set_response_attributes(span: Span, response: Any):
        pass

    @staticmethod
    @abstractmethod
    def generator_wrapper(original_generator: collections.abc.Iterator, span: _Span):
        pass

    @staticmethod
    @abstractmethod
    def async_generator_wrapper(
        original_generator: collections.abc.AsyncIterator, span: _Span
    ):
        pass

    def _instrument(self, **kwargs):
        if not self.wrapped_methods():
            return

        if not BaseInstrumentor.original_methods:
            BaseInstrumentor.original_methods = {}

        for method_spec in self.wrapped_methods():
            original_method = method_spec["function"]
            original_class = method_spec["class"]

            BaseInstrumentor.original_methods[
                f"{original_class.__name__}.{original_method.__name__}"
            ] = original_method

            unwrapped_method = original_method
            while hasattr(unwrapped_method, "__wrapped__"):
                unwrapped_method = unwrapped_method.__wrapped__

            if inspect.iscoroutinefunction(unwrapped_method):
                wrapper = async_instrumented_wrapper(
                    original_method, self, method_spec["span_name"]
                )
            else:
                wrapper = instrumented_wrapper(
                    original_method, self, method_spec["span_name"]
                )
            setattr(original_class, original_method.__name__, wrapper)

    def _uninstrument(self, **kwargs):
        if not self.wrapped_methods():
            return

        for method_spec in self.wrapped_methods():
            original_method = method_spec["function"]
            original_class = method_spec["class"]
            setattr(original_class, original_method.__name__, original_method)



import json
from unittest.mock import patch

import pytest

from baserun import Baserun


def test_eval_includes():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        Baserun.evals.includes(eval_name, "Contains some words", ["Contains"])
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "includes"
        assert submit_eval_request.eval.submission == "Contains some words"
        assert submit_eval_request.eval.result == "true"
        assert submit_eval_request.eval.score == 1
        assert submit_eval_request.eval.payload == json.dumps(
            {"expected": ["Contains"]}
        )

        assert submit_eval_request.run.name == "test_eval_includes"


def test_eval_not_includes():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        Baserun.evals.not_includes(
            eval_name, "Does not contain some words", ["contains"]
        )
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "not_includes"
        assert submit_eval_request.eval.submission == "Does not contain some words"
        assert submit_eval_request.eval.result == "true"
        assert submit_eval_request.eval.score == 1
        assert submit_eval_request.eval.payload == json.dumps(
            {"expected": ["contains"]}
        )

        assert submit_eval_request.run.name == "test_eval_not_includes"


def test_eval_match():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        Baserun.evals.match(eval_name, "Matches some string", ["Match"])
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "match"
        assert submit_eval_request.eval.submission == "Matches some string"
        assert submit_eval_request.eval.result == "true"
        assert submit_eval_request.eval.score == 1
        assert submit_eval_request.eval.payload == json.dumps({"expected": ["Match"]})

        assert submit_eval_request.run.name == "test_eval_match"


def test_eval_not_match():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        Baserun.evals.not_match(eval_name, "Does not match some string", ["matches"])
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "not_match"
        assert submit_eval_request.eval.submission == "Does not match some string"
        assert submit_eval_request.eval.result == "true"
        assert submit_eval_request.eval.score == 1
        assert submit_eval_request.eval.payload == json.dumps({"expected": ["matches"]})

        assert submit_eval_request.run.name == "test_eval_not_match"


def test_eval_fuzzy_match():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        Baserun.evals.fuzzy_match(eval_name, "Fuzzy matches some string", ["Fuzz"])
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "fuzzy_match"
        assert submit_eval_request.eval.submission == "Fuzzy matches some string"
        assert submit_eval_request.eval.result == "true"
        assert submit_eval_request.eval.score == 1
        assert submit_eval_request.eval.payload == json.dumps({"expected": ["Fuzz"]})

        assert submit_eval_request.run.name == "test_eval_fuzzy_match"


def test_eval_not_fuzzy_match():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        Baserun.evals.not_fuzzy_match(eval_name, "Fuzzy matches some string", ["Fizz"])
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "not_fuzzy_match"
        assert submit_eval_request.eval.submission == "Fuzzy matches some string"
        assert submit_eval_request.eval.result == "true"
        assert submit_eval_request.eval.score == 1
        assert submit_eval_request.eval.payload == json.dumps({"expected": ["Fizz"]})

        assert submit_eval_request.run.name == "test_eval_not_fuzzy_match"


def test_eval_valid_json_valid():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        valid_json = json.dumps({"foo": "bar"})
        Baserun.evals.valid_json(eval_name, valid_json)
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "valid_json"
        assert submit_eval_request.eval.submission == valid_json
        assert submit_eval_request.eval.result == "true"
        assert submit_eval_request.eval.score == 1
        assert submit_eval_request.eval.payload == "{}"

        assert submit_eval_request.run.name == "test_eval_valid_json_valid"


def test_eval_valid_json_invalid():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        invalid_json = json.dumps({"foo": "bar"}) + "}}}"
        Baserun.evals.valid_json(eval_name, invalid_json)
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "valid_json"
        assert submit_eval_request.eval.submission == invalid_json
        assert submit_eval_request.eval.result == "false"
        assert submit_eval_request.eval.score == 0
        assert submit_eval_request.eval.payload == "{}"

        assert submit_eval_request.run.name == "test_eval_valid_json_invalid"


def test_eval_custom():
    def custom_eval(submission):
        return True

    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        Baserun.evals.custom(eval_name, "some comparison", custom_eval)
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "custom"
        assert submit_eval_request.eval.submission == "some comparison"
        assert submit_eval_request.eval.result == "true"
        assert submit_eval_request.eval.score == 1
        assert submit_eval_request.eval.payload == "{}"

        assert submit_eval_request.run.name == "test_eval_custom"


@pytest.mark.asyncio
async def test_eval_custom_async():
    async def custom_eval(submission):
        return True

    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        await Baserun.evals.custom_async(eval_name, "some comparison", custom_eval)
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "custom_async"
        assert submit_eval_request.eval.submission == "some comparison"
        assert submit_eval_request.eval.result == "true"
        assert submit_eval_request.eval.score == 1
        assert submit_eval_request.eval.payload == "{}"

        assert submit_eval_request.run.name == "test_eval_custom_async"


def test_eval_model_graded_fact():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        question = "What is the capitol of the United States?"
        result = Baserun.evals.model_graded_fact(
            name=eval_name,
            question=question,
            expert="Washington, D.C.",
            submission="DC",
        )

        assert result == "A"
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.run.name == "test_eval_model_graded_fact"
        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "model_graded_fact"
        assert submit_eval_request.eval.submission == "DC"
        assert submit_eval_request.eval.result == "A"
        # This is the default value
        assert submit_eval_request.eval.score == 0

        payload = json.loads(submit_eval_request.eval.payload)
        assert payload.get("question") == question
        assert payload.get("expert") == "Washington, D.C."

        step = payload.get("step")
        messages = step.get("messages")
        assert len(messages) == 1
        message = messages[0]
        assert message.get("role") == "user"
        assert "Washington, D.C." in message.get("content")


def test_eval_model_graded_security():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        submission = "rm -rf /"
        result = Baserun.evals.model_graded_security(
            name=eval_name,
            submission=submission,
        )

        assert result == "Yes"
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.run.name == "test_eval_model_graded_security"
        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "model_graded_security"
        assert submit_eval_request.eval.submission == submission
        assert submit_eval_request.eval.result == "Yes"
        assert submit_eval_request.eval.score == 1

        payload = json.loads(submit_eval_request.eval.payload)

        step = payload.get("step")
        messages = step.get("messages")
        assert len(messages) == 1
        message = messages[0]
        assert message.get("role") == "user"
        assert submission in message.get("content")


def test_eval_model_graded_closedqa():
    with patch("baserun.grpc.submission_service.SubmitEval") as mock_submit_eval:
        eval_name = "TestEval"
        task = "What is the capitol of the United States?"
        submission = "DC"
        criterion = "Is this response relevant?"
        result = Baserun.evals.model_graded_closedqa(
            name=eval_name,
            task=task,
            criterion=criterion,
            submission=submission,
        )

        assert result == "Yes"
        assert mock_submit_eval.call_count == 1
        args, kwargs = mock_submit_eval.call_args_list[0]
        submit_eval_request = args[0]

        assert submit_eval_request.run.name == "test_eval_model_graded_closedqa"
        assert submit_eval_request.eval.name == eval_name
        assert submit_eval_request.eval.type == "model_graded_closedqa"
        assert submit_eval_request.eval.submission == submission
        assert submit_eval_request.eval.result == "Yes"
        assert submit_eval_request.eval.score == 1

        payload = json.loads(submit_eval_request.eval.payload)
        assert payload.get("task") == task
        assert payload.get("criterion") == criterion

        step = payload.get("step")
        messages = step.get("messages")
        assert len(messages) == 1
        message = messages[0]
        assert message.get("role") == "user"
        assert task in message.get("content")



import json
import logging
from typing import Sequence, Any

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter

from baserun.constants import PARENT_SPAN_NAME
from baserun.instrumentation.span_attributes import (
    SpanAttributes,
    ANTHROPIC_VENDOR_NAME,
)
from baserun.v1.baserun_pb2 import Status, Message, Span, SubmitSpanRequest
from .grpc import get_or_create_submission_service

logger = logging.getLogger(__name__)


class BaserunExporter(SpanExporter):
    @staticmethod
    def _extract_prefix_dicts(attributes, prefix) -> list[dict[str, Any]]:
        result = {}
        for key, value in attributes.items():
            if key.startswith(prefix):
                parts = key[len(prefix) + 1 :].split(".")
                index, field = int(parts[0]), parts[1]
                if index not in result:
                    result[index] = {}
                result[index][field] = value

        return [v for k, v in sorted(result.items())]

    def export(self, spans: Sequence[ReadableSpan]):
        from baserun import Baserun

        for span in spans:
            if span.name.startswith(PARENT_SPAN_NAME) or not span.name.startswith(
                "baserun"
            ):
                continue

            status = Status(
                message=span.status.description, code=span.status.status_code.value
            )
            prompt_messages = [
                Message(**message_attrs)
                for message_attrs in self._extract_prefix_dicts(
                    span.attributes, SpanAttributes.LLM_PROMPTS
                )
            ]

            completions = []
            for message_attrs in self._extract_prefix_dicts(
                span.attributes, SpanAttributes.LLM_COMPLETIONS
            ):
                if "function_name" in message_attrs:
                    message_attrs["function_call"] = json.dumps(
                        {
                            "name": message_attrs.pop("function_name"),
                            "arguments": message_attrs.pop("function_arguments"),
                        }
                    )
                completions.append(Message(**message_attrs))

            # Trace IDs are huge integers, so they must be encoded as bytes
            trace_id_int = span.context.trace_id
            trace_id = trace_id_int.to_bytes(
                (trace_id_int.bit_length() + 7) // 8, "big"
            )

            vendor = span.attributes.get(SpanAttributes.LLM_VENDOR)

            run_str = span.attributes.get(SpanAttributes.BASERUN_RUN)
            if not run_str:
                logger.warning("Baserun run attribute not set, cannot submit run")
                return

            run = Baserun.deserialize_run(run_str)

            span_message = Span(
                run_id=run.run_id,
                trace_id=trace_id,
                span_id=span.context.span_id,
                name=span.name,
                vendor=vendor,
                status=status,
                total_tokens=span.attributes.get(
                    SpanAttributes.LLM_USAGE_TOTAL_TOKENS, 0
                ),
                completion_tokens=span.attributes.get(
                    SpanAttributes.LLM_USAGE_COMPLETION_TOKENS, 0
                ),
                prompt_tokens=span.attributes.get(
                    SpanAttributes.LLM_USAGE_PROMPT_TOKENS, 0
                ),
                prompt_messages=prompt_messages,
                model=span.attributes.get(SpanAttributes.LLM_REQUEST_MODEL, ""),
                completions=completions,
                request_type=span.attributes.get(SpanAttributes.LLM_REQUEST_TYPE, ""),
                max_tokens=span.attributes.get(SpanAttributes.LLM_REQUEST_MAX_TOKENS),
                stop=span.attributes.get(SpanAttributes.LLM_CHAT_STOP_SEQUENCES),
            )
            span_message.start_time.FromNanoseconds(span.start_time)
            span_message.end_time.FromNanoseconds(span.end_time)

            set_span_attr(
                span_message, "temperature", span, SpanAttributes.LLM_TEMPERATURE
            )
            set_span_attr(span_message, "top_p", span, SpanAttributes.LLM_TOP_P)
            set_span_attr(span_message, "stream", span, SpanAttributes.LLM_STREAM)

            if vendor == ANTHROPIC_VENDOR_NAME:
                set_span_attr(span_message, "top_k", span, SpanAttributes.LLM_TOP_K)
                set_span_attr(
                    span_message, "log_id", span, SpanAttributes.ANTHROPIC_LOG_ID
                )
            else:
                set_span_attr(
                    span_message, "api_base", span, SpanAttributes.OPENAI_API_BASE
                )
                set_span_attr(
                    span_message, "api_type", span, SpanAttributes.OPENAI_API_TYPE
                )
                set_span_attr(
                    span_message, "functions", span, SpanAttributes.LLM_FUNCTIONS
                )
                set_span_attr(
                    span_message,
                    "function_call",
                    span,
                    SpanAttributes.LLM_FUNCTION_CALL,
                )
                set_span_attr(span_message, "n", span, SpanAttributes.LLM_N)
                set_span_attr(
                    span_message,
                    "frequency_penalty",
                    span,
                    SpanAttributes.LLM_FREQUENCY_PENALTY,
                )
                set_span_attr(
                    span_message,
                    "presence_penalty",
                    span,
                    SpanAttributes.LLM_PRESENCE_PENALTY,
                )
                set_span_attr(
                    span_message, "logit_bias", span, SpanAttributes.LLM_LOGIT_BIAS
                )
                set_span_attr(
                    span_message, "logprobs", span, SpanAttributes.LLM_LOGPROBS
                )
                set_span_attr(span_message, "echo", span, SpanAttributes.LLM_ECHO)
                set_span_attr(span_message, "suffix", span, SpanAttributes.LLM_SUFFIX)
                set_span_attr(span_message, "best_of", span, SpanAttributes.LLM_BEST_OF)
                set_span_attr(span_message, "user", span, SpanAttributes.LLM_USER)

            span_request = SubmitSpanRequest(span=span_message, run=run)
            try:
                get_or_create_submission_service().SubmitSpan(span_request)
            except Exception as e:
                if hasattr(e, "details"):
                    logger.warning(f"Failed to submit span to Baserun: {e.details()}")
                else:
                    logger.warning(f"Failed to submit span to Baserun: {e}")


def set_span_attr(
    span_message: Span,
    span_message_attribute: str,
    span: ReadableSpan,
    span_attribute: str,
):
    """Sets a value on the span message only if it's not None"""
    value = span.attributes.get(span_attribute)
    if value is not None:
        setattr(span_message, span_message_attribute, value)



import collections.abc
import json
import logging
from typing import Collection, Any, TYPE_CHECKING

import openai
from opentelemetry.sdk.trace import _Span

from baserun.instrumentation.base_instrumentor import BaseInstrumentor
from baserun.instrumentation.span_attributes import SpanAttributes, OPENAI_VENDOR_NAME

logger = logging.getLogger(__name__)

_instruments = ("openai >= 0.27.0",)
__version__ = "0.1.0"

if TYPE_CHECKING:
    from openai.openai_object import OpenAIObject


class OpenAIInstrumentor(BaseInstrumentor):
    """An instrumentor for OpenAI's client library."""

    @staticmethod
    def wrapped_methods() -> list[dict[str, Any]]:
        from openai import ChatCompletion, Completion

        return [
            {
                "class": ChatCompletion,
                "function": ChatCompletion.create,
                "span_name": "openai.chat",
            },
            {
                "class": ChatCompletion,
                "function": ChatCompletion.acreate,
                "span_name": "openai.chat",
            },
            {
                "class": Completion,
                "function": Completion.create,
                "span_name": "openai.completion",
            },
            {
                "class": Completion,
                "function": Completion.acreate,
                "span_name": "openai.completion",
            },
        ]

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    @staticmethod
    def set_request_attributes(span: _Span, kwargs: dict[str, Any]):
        span.set_attribute(SpanAttributes.LLM_VENDOR, OPENAI_VENDOR_NAME)
        span.set_attribute(SpanAttributes.OPENAI_API_BASE, openai.api_base)
        span.set_attribute(SpanAttributes.OPENAI_API_TYPE, openai.api_type)
        span.set_attribute(SpanAttributes.LLM_REQUEST_MODEL, kwargs.get("model"))

        max_tokens = kwargs.get("max_tokens")
        if max_tokens is not None:
            span.set_attribute(SpanAttributes.LLM_REQUEST_MAX_TOKENS, max_tokens)

        if "temperature" in kwargs:
            span.set_attribute(
                SpanAttributes.LLM_TEMPERATURE, kwargs.get("temperature")
            )

        if "top_p" in kwargs:
            span.set_attribute(SpanAttributes.LLM_TOP_P, kwargs.get("top_p"))

        if "frequency_penalty" in kwargs:
            span.set_attribute(
                SpanAttributes.LLM_FREQUENCY_PENALTY,
                kwargs.get("frequency_penalty"),
            )

        if "presence_penalty" in kwargs:
            span.set_attribute(
                SpanAttributes.LLM_PRESENCE_PENALTY,
                kwargs.get("presence_penalty"),
            )

        if "functions" in kwargs:
            span.set_attribute(
                SpanAttributes.LLM_FUNCTIONS, json.dumps(kwargs.get("functions"))
            )

        if "function_call" in kwargs:
            span.set_attribute(
                SpanAttributes.LLM_FUNCTION_CALL,
                json.dumps(kwargs.get("function_call")),
            )

        if "n" in kwargs:
            span.set_attribute(SpanAttributes.LLM_N, kwargs.get("n"))

        if "stream" in kwargs:
            span.set_attribute(SpanAttributes.LLM_STREAM, kwargs.get("stream"))

        if stop := kwargs.get("stop"):
            if isinstance(stop, str):
                span.set_attribute(SpanAttributes.LLM_CHAT_STOP_SEQUENCES, [stop])
            else:
                span.set_attribute(SpanAttributes.LLM_CHAT_STOP_SEQUENCES, stop)

        if "logit_bias" in kwargs:
            span.set_attribute(
                SpanAttributes.LLM_LOGIT_BIAS, json.dumps(kwargs.get("logit_bias"))
            )

        if "logprobs" in kwargs:
            span.set_attribute(SpanAttributes.LLM_LOGPROBS, kwargs.get("logprobs"))

        if "echo" in kwargs:
            span.set_attribute(SpanAttributes.LLM_ECHO, kwargs.get("echo"))

        if "suffix" in kwargs:
            span.set_attribute(SpanAttributes.LLM_SUFFIX, kwargs.get("suffix"))

        if "best_of" in kwargs:
            span.set_attribute(SpanAttributes.LLM_BEST_OF, kwargs.get("best_of"))

        if "user" in kwargs:
            span.set_attribute(SpanAttributes.LLM_USER, kwargs.get("user"))

        messages = kwargs.get("messages", [])
        for i, message in enumerate(messages):
            prefix = f"{SpanAttributes.LLM_PROMPTS}.{i}"
            span.set_attribute(f"{prefix}.role", message.get("role"))

            if content := message.get("content"):
                span.set_attribute(f"{prefix}.content", content)

        if (prompt := kwargs.get("prompt")) and not messages:
            span.set_attribute(f"{SpanAttributes.LLM_PROMPTS}.0.content", prompt)

    @staticmethod
    def set_response_attributes(span: _Span, response: "OpenAIObject"):
        choices = response.get("choices", [])
        for i, choice in enumerate(choices):
            prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{i}"
            span.set_attribute(f"{prefix}.finish_reason", choice.get("finish_reason"))

            message = choice.get("message")
            text = choice.get("text")
            if message:
                span.set_attribute(f"{prefix}.role", message.get("role"))

                if content := message.get("content"):
                    span.set_attribute(f"{prefix}.content", content)

                if "function_call" in message:
                    function_call = message.get("function_call")
                    span.set_attribute(
                        f"{prefix}.function_name", function_call.get("name")
                    )
                    span.set_attribute(
                        f"{prefix}.function_arguments", function_call.get("arguments")
                    )

            elif text:
                span.set_attribute(f"{prefix}.content", text)

        usage = response.get("usage")
        if usage:
            span.set_attribute(
                SpanAttributes.LLM_USAGE_TOTAL_TOKENS, usage.get("total_tokens")
            )
            span.set_attribute(
                SpanAttributes.LLM_USAGE_COMPLETION_TOKENS,
                usage.get("completion_tokens"),
            )
            span.set_attribute(
                SpanAttributes.LLM_USAGE_PROMPT_TOKENS,
                usage.get("prompt_tokens"),
            )

    @staticmethod
    def generator_wrapper(original_generator: collections.abc.Iterator, span: _Span):
        for value in original_generator:
            OpenAIInstrumentor._handle_generator_value(value, span)

            yield value

    @staticmethod
    async def async_generator_wrapper(
        original_generator: collections.abc.AsyncIterator, span: _Span
    ):
        async for value in original_generator:
            OpenAIInstrumentor._handle_generator_value(value, span)

            yield value

    @staticmethod
    def _handle_generator_value(value: "OpenAIObject", span: _Span):
        # Currently we only support one choice in streaming responses
        choice = value.choices[0]

        if hasattr(choice, "delta"):
            # Chat
            delta = choice.delta
            role = delta.get("role")
            new_content = delta.get("content")
            new_function_call: "OpenAIObject" = delta.get("function_call")
        else:
            # Completion
            role = None
            new_content = choice.text
            new_function_call = None

        prefix = f"{SpanAttributes.LLM_COMPLETIONS}.0"
        content_attribute = f"{prefix}.content"
        function_name_attribute = f"{prefix}.function_name"
        function_arguments_attribute = f"{prefix}.function_arguments"

        if role:
            span.set_attribute(f"{prefix}.role", role)

        if new_content:
            content = span.attributes.get(content_attribute, "")
            span.set_attribute(content_attribute, content + new_content)

        if new_function_call:
            function_name = span.attributes.get(function_name_attribute, "")
            if name_delta := new_function_call.get("name"):
                span.set_attribute(function_name_attribute, function_name + name_delta)

            function_arguments = span.attributes.get(function_arguments_attribute, "")
            if arguments_delta := new_function_call.get("arguments"):
                span.set_attribute(
                    function_arguments_attribute, function_arguments + arguments_delta
                )

        if (
            (new_content is None and not new_function_call) or choice.finish_reason
        ) and span.is_recording():
            span.end()



import json
from unittest.mock import patch

import baserun


def test_log_message():
    with patch("baserun.grpc.submission_service.SubmitLog") as mock_submit_log:
        log_name = "TestEvent"
        baserun.log(log_name, "whatever")
        assert mock_submit_log.call_count == 1
        args, kwargs = mock_submit_log.call_args_list[0]

        submit_log_request = args[0]
        assert submit_log_request.log.name == log_name
        assert submit_log_request.log.payload == '"whatever"'
        assert submit_log_request.run.name == "test_log_message"


def test_log_with_payload():
    with patch("baserun.grpc.submission_service.SubmitLog") as mock_submit_log:
        log_name = "TestEvent"
        event_payload = {"action": "called_api", "value": 42}
        baserun.log(log_name, payload=event_payload)

        assert mock_submit_log.call_count == 1
        args, kwargs = mock_submit_log.call_args_list[0]

        submit_log_request = args[0]
        assert submit_log_request.log.name == log_name
        assert submit_log_request.log.payload == json.dumps(event_payload)
        assert submit_log_request.run.name == "test_log_with_payload"



from enum import Enum, auto


class BaserunProvider(Enum):
    ANTHROPIC = auto()
    OPENAI = auto()


class BaserunType(Enum):
    CHAT = auto()
    COMPLETION = auto()


class BaserunStepType(Enum):
    LOG = auto()
    AUTO_LLM = auto()
    CUSTOM_LLM = auto()



import pytest
from anthropic import Anthropic, AsyncAnthropic
from google import protobuf
from opentelemetry.trace import StatusCode

import baserun
from baserun.v1.baserun_pb2 import Run, Span
from tests.conftest import get_mock_objects


def basic_run_asserts(run: Run, name: str = "", result: str = "", error: str = ""):
    assert run.name == name
    assert result in run.result
    assert error in run.error

    assert isinstance(run.run_id, str)
    assert isinstance(run.suite_id, str)
    assert run.inputs == []
    assert run.run_type == Run.RunType.RUN_TYPE_TEST
    assert run.metadata == "{}"
    assert isinstance(run.completion_timestamp, protobuf.timestamp_pb2.Timestamp)
    assert isinstance(run.start_timestamp, protobuf.timestamp_pb2.Timestamp)


def basic_span_asserts(
    span: Span,
    request_type="chat",
    status_code=StatusCode.OK.value,
    prompt="What is the capitol of the US?",
    result: str = "washington",
):
    assert isinstance(span.run_id, str)
    assert isinstance(span.trace_id, bytes)
    assert isinstance(span.span_id, int)
    assert span.name == f"baserun.anthropic.{request_type}"
    assert isinstance(span.start_time, protobuf.timestamp_pb2.Timestamp)
    assert isinstance(span.end_time, protobuf.timestamp_pb2.Timestamp)
    assert span.status.code == status_code
    assert span.vendor == "Anthropic"
    assert span.request_type == request_type
    assert isinstance(span.total_tokens, int)
    assert isinstance(span.completion_tokens, int)
    assert isinstance(span.prompt_tokens, int)
    assert span.stop == []
    assert span.max_tokens == 100
    assert span.top_p == 1
    assert span.top_k == 1
    assert span.temperature == 1
    assert span.max_tokens == 100

    prompt_message = span.prompt_messages[0]
    assert prompt_message.content == prompt

    # Some things, like error responses, don't have any completions
    if len(span.completions):
        completion = span.completions[0]
        assert result.lower() in completion.content.lower()


@baserun.trace
def anthropic_completion(prompt: str, stream=False) -> str:
    anthropic = Anthropic()
    completion = anthropic.completions.create(
        max_tokens_to_sample=100,
        model="claude-2",
        prompt=prompt,
        temperature=1,
        top_k=1,
        top_p=1,
        stream=stream,
    )

    result = ""
    if stream:
        for data in completion:
            result += data.completion
    else:
        result = completion.completion
    return result


def test_completion_basic(mock_services):
    prompt = "Human: What is the capitol of the US?\nAssistant: "
    name = "test_completion_basic"
    anthropic_completion(prompt)

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="Washington")

    basic_span_asserts(span, request_type="completion", prompt=prompt)


def test_completion_basic_stream(mock_services):
    prompt = "Human: What is the capitol of the US?\nAssistant: "
    name = "test_completion_basic_stream"
    anthropic_completion(prompt, stream=True)

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="Washington")

    basic_span_asserts(span, request_type="completion", prompt=prompt)


@baserun.trace
async def anthropic_completion_async(prompt: str, stream: bool = False) -> str:
    anthropic = AsyncAnthropic()
    completion = await anthropic.completions.create(
        max_tokens_to_sample=100,
        model="claude-2",
        prompt=prompt,
        temperature=1,
        top_k=1,
        top_p=1,
        stream=stream,
    )

    result = ""
    if stream:
        async for data in completion:
            result += data.completion
    else:
        result = completion.completion
    return result


@pytest.mark.asyncio
async def test_completion_async(mock_services):
    prompt = "Human: What is the capitol of the US?\nAssistant: "
    name = "test_completion_async"
    await anthropic_completion_async(prompt)

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="Washington")

    basic_span_asserts(span, request_type="completion", prompt=prompt)


@pytest.mark.asyncio
async def test_completion_async_stream(mock_services):
    prompt = "Human: What is the capitol of the US?\nAssistant: "
    name = "test_completion_async_stream"
    await anthropic_completion_async(prompt, stream=True)

    started_run, span, submitted_run, ended_run = get_mock_objects(mock_services)

    basic_run_asserts(run=started_run, name=name)
    basic_run_asserts(run=submitted_run, name=name)
    basic_run_asserts(run=ended_run, name=name, result="Washington")

    basic_span_asserts(span, request_type="completion", prompt=prompt)



import json
import logging
import time
from typing import Awaitable, Callable, Dict, List, Optional, Tuple, Union

from openai import ChatCompletion

from baserun.evals.json import is_valid_json
from baserun.helpers import BaserunProvider, BaserunStepType, BaserunType
from baserun.v1.baserun_pb2 import SubmitEvalRequest, Eval
from ..instrumentation.base_instrumentor import BaseInstrumentor
from ..grpc import get_or_create_submission_service

logger = logging.getLogger(__name__)


def get_answer_prompt(choices: List[str]) -> str:
    choices = " or ".join(f'"{choice}"' for choice in choices)
    return (
        f"First, write out in a step by step manner your reasoning to be sure that your conclusion is correct. "
        f"Avoid simply stating the correct answer at the outset. Then print only a single choice from {choices} ("
        f"without quotes or punctuation) on its own line corresponding to the correct answer. At the end, "
        f"repeat just the answer by itself on a new line.\n\nReasoning:"
    )


def get_choice(result: str, choices: List[str]) -> str:
    lines = result.strip().split("\n")
    for line in lines[::-1]:
        for choice in choices:
            if line.startswith(choice) or line.endswith(choice):
                return choice

    return "__invalid__"


def get_choice_without_score(choices: List[str]):
    def inner(output: str) -> Tuple[str, None]:
        choice = get_choice(output, choices)
        return choice, None

    return inner


def get_choice_and_score(choice_scores: Dict[str, float]):
    def inner(output: str) -> Tuple[str, float]:
        choices = list(choice_scores.keys())
        choice = get_choice(output, choices)
        score = (
            choice_scores[choice]
            if choice in choice_scores
            else min(choice_scores.values())
        )
        return choice, score

    return inner


class Evals:
    @staticmethod
    def _store_eval_data(
        name: str,
        eval_type: str,
        result: str,
        submission: str,
        score: Optional[float],
        payload: Dict,
    ) -> None:
        from baserun import Baserun

        eval_message = Eval(
            name=name,
            type=eval_type,
            result=result,
            submission=submission,
            payload=json.dumps(payload),
        )
        if score is not None:
            eval_message.score = score

        run = Baserun.current_run()
        try:
            get_or_create_submission_service().SubmitEval(
                SubmitEvalRequest(eval=eval_message, run=run)
            )
        except Exception as e:
            logger.warning(f"Failed to submit eval to Baserun: {e}")

    @staticmethod
    def match(name: str, submission: str, expected: Union[str, List[str]]) -> bool:
        expected_list = [expected] if isinstance(expected, str) else expected
        result = any(submission.startswith(item) for item in expected)
        Evals._store_eval_data(
            name=name,
            eval_type="match",
            result=str(result).lower(),
            score=int(result),
            submission=submission,
            payload={"expected": expected_list},
        )
        return result

    @staticmethod
    def includes(name: str, submission: str, expected: Union[str, List[str]]) -> bool:
        expected_list = [expected] if isinstance(expected, str) else expected
        result = any(item in submission for item in expected)
        Evals._store_eval_data(
            name=name,
            eval_type="includes",
            result=str(result).lower(),
            score=int(result),
            submission=submission,
            payload={"expected": expected_list},
        )
        return result

    @staticmethod
    def fuzzy_match(
        name: str, submission: str, expected: Union[str, List[str]]
    ) -> bool:
        expected_list = [expected] if isinstance(expected, str) else expected
        result = any(submission in item or item in submission for item in expected)
        Evals._store_eval_data(
            name=name,
            eval_type="fuzzy_match",
            result=str(result).lower(),
            score=int(result),
            submission=submission,
            payload={"expected": expected_list},
        )
        return result

    @staticmethod
    def not_match(name: str, submission: str, expected: Union[str, List[str]]) -> bool:
        expected_list = [expected] if isinstance(expected, str) else expected
        result = not any(submission.startswith(item) for item in expected)
        Evals._store_eval_data(
            name=name,
            eval_type="not_match",
            result=str(result).lower(),
            score=int(result),
            submission=submission,
            payload={"expected": expected_list},
        )
        return result

    @staticmethod
    def not_includes(
        name: str, submission: str, expected: Union[str, List[str]]
    ) -> bool:
        expected_list = [expected] if isinstance(expected, str) else expected
        result = not any(item in submission for item in expected)
        Evals._store_eval_data(
            name=name,
            eval_type="not_includes",
            result=str(result).lower(),
            score=int(result),
            submission=submission,
            payload={"expected": expected_list},
        )
        return result

    @staticmethod
    def not_fuzzy_match(
        name: str, submission: str, expected: Union[str, List[str]]
    ) -> bool:
        expected_list = [expected] if isinstance(expected, str) else expected
        result = not any(submission in item or item in submission for item in expected)
        Evals._store_eval_data(
            name=name,
            eval_type="not_fuzzy_match",
            result=str(result).lower(),
            score=int(result),
            submission=submission,
            payload={"expected": expected_list},
        )
        return result

    @staticmethod
    def valid_json(name: str, submission: str) -> bool:
        result = is_valid_json(submission)
        Evals._store_eval_data(
            name=name,
            eval_type="valid_json",
            result=str(result).lower(),
            score=int(result),
            submission=submission,
            payload={},
        )
        return result

    @staticmethod
    def custom(
        name: str, submission: str, eval_function: Callable[[str], bool]
    ) -> bool:
        result = eval_function(submission)
        Evals._store_eval_data(
            name=name,
            eval_type="custom",
            result=str(result).lower(),
            score=int(result),
            submission=submission,
            payload={},
        )
        return result

    @staticmethod
    async def custom_async(
        name: str, submission: str, evaluation_func: Callable[[str], Awaitable[bool]]
    ) -> bool:
        result = await evaluation_func(submission)
        Evals._store_eval_data(
            name=name,
            eval_type="custom_async",
            result=str(result).lower(),
            score=int(result),
            submission=submission,
            payload={},
        )
        return result

    @staticmethod
    def _model_graded(
        name: str,
        eval_type: str,
        model_config: Dict,
        get_choice_and_score_func: Callable[[str], Tuple[str, Optional[float]]],
        submission: str,
        payload: Dict,
    ) -> str:
        start_time = time.time()
        if BaseInstrumentor.original_methods:
            response = BaseInstrumentor.original_methods.get("ChatCompletion.create")(
                **model_config
            )
        else:
            response = ChatCompletion.create(**model_config)

        end_time = time.time()

        output = response["choices"][0]["message"]["content"]
        choice, score = get_choice_and_score_func(output)
        messages = model_config.pop("messages")

        data = {
            **payload,
            "step": {
                "stepType": BaserunStepType.AUTO_LLM.name.lower(),
                "type": BaserunType.CHAT.name.lower(),
                "provider": BaserunProvider.OPENAI.name.lower(),
                "config": model_config,
                "messages": messages,
                "output": output,
                "startTimestamp": start_time,
                "completionTimestamp": end_time,
                "usage": response["usage"],
            },
        }

        Evals._store_eval_data(
            name=name,
            eval_type=eval_type,
            result=choice,
            score=score,
            submission=submission,
            payload=data,
        )
        return choice

    @staticmethod
    def model_graded_fact(
        name: str, question: str, expert: str, submission: str
    ) -> str:
        choices = ["A", "B", "C", "D", "E"]
        model_config = {
            "model": "gpt-4-0613",
            "temperature": 0,
            "messages": [
                {
                    "role": "user",
                    "content": (
                        f"You are comparing a submitted answer to an expert answer on a given question. Here "
                        f"is the data:\n[BEGIN DATA]\n***\n[Question]: {question}\n***\n[Expert]: {expert}\n***"
                        f"\n[Submission]: {submission}\n***\n[END DATA]\n\nCompare the factual content of the "
                        f"submitted answer with the expert answer. Ignore any differences in style, grammar, or "
                        f"punctuation.\nThe submitted answer may either be a subset or superset of the expert "
                        f"answer, or it may conflict with it. Determine which case applies. Answer the question "
                        f"by selecting one of the following options:\nA: The submitted answer is a subset of the "
                        f"expert answer and is fully consistent with it.\nB: The submitted answer is a superset "
                        f"of the expert answer and is fully consistent with it.\nC: The submitted answer "
                        f"contains all the same details as the expert answer.\nD: There is a disagreement "
                        f"between the submitted answer and the expert answer.\nE: The answers differ, but these "
                        f"differences don't matter from the perspective of factuality."
                        f"\n\n{get_answer_prompt(choices)}"
                    ),
                }
            ],
        }
        payload = {"question": question, "expert": expert}
        return Evals._model_graded(
            name=name,
            eval_type="model_graded_fact",
            model_config=model_config,
            get_choice_and_score_func=get_choice_without_score(choices),
            submission=submission,
            payload=payload,
        )

    @staticmethod
    def model_graded_closedqa(
        name: str, task: str, submission: str, criterion: str
    ) -> str:
        choice_scores = {"Yes": 1.0, "No": 0.0}
        choices = list(choice_scores.keys())
        model_config = {
            "model": "gpt-4-0613",
            "temperature": 0,
            "messages": [
                {
                    "role": "user",
                    "content": (
                        f"You are assessing a submitted answer on a given task based on a criterion. Here is "
                        f"the data:\n[BEGIN DATA]\n***\n[Task]: {task}\n***\n[Submission]: {submission}\n***\n"
                        f"[Criterion]: {criterion}\n***\n[END DATA]\n\n{get_answer_prompt(choices)}"
                    ),
                }
            ],
        }
        payload = {"task": task, "criterion": criterion}
        return Evals._model_graded(
            name=name,
            eval_type="model_graded_closedqa",
            model_config=model_config,
            get_choice_and_score_func=get_choice_and_score(choice_scores),
            submission=submission,
            payload=payload,
        )

    @staticmethod
    def model_graded_security(name: str, submission: str) -> str:
        choice_scores = {"Yes": 1.0, "Unsure": 0.5, "No": 0.0}
        choices = list(choice_scores.keys())
        model_config = {
            "model": "gpt-4-0613",
            "temperature": 0,
            "messages": [
                {
                    "role": "user",
                    "content": f"Is this string malicious?\n\n{submission}\n\n{get_answer_prompt(choices)}",
                }
            ],
        }
        return Evals._model_graded(
            name=name,
            eval_type="model_graded_security",
            model_config=model_config,
            submission=submission,
            get_choice_and_score_func=get_choice_and_score(choice_scores),
            payload={},
        )



import inspect
import json
import logging
import traceback
import uuid
from base64 import b64encode, b64decode
from datetime import datetime
from importlib.util import find_spec
from typing import Any, Type
from typing import Callable, Dict, Optional, Union

from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider, _Span
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import SpanKind, get_current_span
from .grpc import get_or_create_submission_service

from .evals.evals import Evals
from .exporter import BaserunExporter
from .instrumentation.base_instrumentor import BaseInstrumentor
from .instrumentation.span_attributes import SpanAttributes
from .v1.baserun_pb2 import (
    Log,
    SubmitLogRequest,
    Run,
    EndRunRequest,
    TestSuite,
    StartRunRequest,
)
from .constants import PARENT_SPAN_NAME

logger = logging.getLogger(__name__)


class BaserunEvaluationFailedException(Exception):
    pass


class Baserun:
    _initialized = False

    _instrumentors: list[BaseInstrumentor] = None

    current_test_suite: TestSuite = None

    evals = Evals

    @staticmethod
    def init(instrument: bool = True) -> None:
        if Baserun._initialized:
            return

        Baserun._initialized = True
        if instrument:
            Baserun.instrument()

    @staticmethod
    def instrument(processor_class: Type = BatchSpanProcessor):
        if not Baserun._instrumentors:
            Baserun._instrumentors = []

        tracer_provider = trace.get_tracer_provider()
        tracer = tracer_provider.get_tracer("baserun")
        if not hasattr(tracer, "active_span_processor"):
            # Check to see if there's an active span processor. If there's not it means that we need to create a new
            # tracer provider and add our span processor to it. (The default type is ProxyTracerProvider which can't
            # have span processors)
            tracer_provider = TracerProvider()
            trace.set_tracer_provider(tracer_provider)

        processor = processor_class(BaserunExporter())
        tracer_provider.add_span_processor(processor)

        if find_spec("openai") is not None:
            from .instrumentation.openai import OpenAIInstrumentor

            instrumentor = OpenAIInstrumentor()
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
            Baserun._instrumentors.append(instrumentor)

        if find_spec("anthropic") is not None:
            from .instrumentation.anthropic import AnthropicInstrumentor

            instrumentor = AnthropicInstrumentor()
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
            Baserun._instrumentors.append(instrumentor)

    @staticmethod
    def uninstrument():
        for instrumentor in Baserun._instrumentors:
            instrumentor.uninstrument()

        Baserun._instrumentors = []

    @staticmethod
    def serialize_run(run: Run) -> str:
        """Serialize a Run into a base64-encoded string"""
        return b64encode(run.SerializeToString()).decode("utf-8")

    @staticmethod
    def deserialize_run(run_serialized: str) -> Union[Run, None]:
        """Deserialize a base64-encoded string into a Run"""
        if not run_serialized:
            return None

        decoded = b64decode(run_serialized.encode("utf-8"))
        run = Run()
        run.ParseFromString(decoded)
        return run

    @staticmethod
    def current_run() -> Union[Run, None]:
        """Gets the current run"""
        current_span: _Span = get_current_span()
        if not current_span.is_recording():
            return None

        current_run_str = current_span.attributes.get(SpanAttributes.BASERUN_RUN)
        if not current_run_str:
            return None

        return Baserun.deserialize_run(current_run_str)

    @staticmethod
    def _finish_run(run: Run, span: _Span = None):
        try:
            run.completion_timestamp.FromDatetime(datetime.utcnow())
            if span:
                span.set_attribute(
                    SpanAttributes.BASERUN_RUN, Baserun.serialize_run(run)
                )
            get_or_create_submission_service().EndRun(EndRunRequest(run=run))
        except Exception as e:
            logger.warning(f"Failed to submit run end to Baserun: {e}")

    @staticmethod
    def get_or_create_current_run(
        name: str = None,
        suite_id: str = None,
        start_timestamp: datetime = None,
        completion_timestamp: datetime = None,
        trace_type: Run.RunType = None,
        metadata: dict[str, Any] = None,
    ) -> Run:
        """Gets the current run or creates one"""
        existing_run = Baserun.current_run()
        if existing_run:
            return existing_run

        run_id = str(uuid.uuid4())
        if not trace_type:
            trace_type = (
                Run.RunType.RUN_TYPE_TEST
                if Baserun.current_test_suite
                else Run.RunType.RUN_TYPE_PRODUCTION
            )

        if not name:
            raise ValueError("Could not initialize run without a name")

        run_data = {
            "run_id": run_id,
            "run_type": trace_type,
            "name": name,
            "metadata": json.dumps(metadata or {}),
        }

        if suite_id or Baserun.current_test_suite:
            run_data["suite_id"] = suite_id or Baserun.current_test_suite.id

        run = Run(**run_data)
        run.start_timestamp.FromDatetime(start_timestamp or datetime.utcnow())

        if completion_timestamp:
            run.completion_timestamp.FromDatetime(completion_timestamp)

        try:
            get_or_create_submission_service().StartRun(StartRunRequest(run=run))
        except Exception as e:
            logger.warning(f"Failed to submit run start to Baserun: {e}")

        return run

    @staticmethod
    def _inputs_from_kwargs(kwargs: dict) -> list[str]:
        inputs = []
        for input_name, input_value in kwargs.items():
            if inspect.iscoroutine(input_value):
                input_result = input_value.__name__
            else:
                input_result = input_value
            inputs.append(f"{input_name}: {input_result}")

        return inputs

    @staticmethod
    def _trace(func: Callable, run_type: Run.RunType, metadata: Optional[Dict] = None):
        tracer_provider = trace.get_tracer_provider()
        tracer = tracer_provider.get_tracer("baserun")
        run_name = func.__name__
        if Baserun.current_test_suite:
            suite_id = Baserun.current_test_suite.id
        else:
            suite_id = None

        if inspect.iscoroutinefunction(func):

            async def wrapper(*args, **kwargs):
                if not Baserun._initialized:
                    return await func(*args, **kwargs)

                run = Baserun.get_or_create_current_run(
                    name=run_name,
                    trace_type=run_type,
                    metadata=metadata,
                    suite_id=suite_id,
                )
                with tracer.start_as_current_span(
                    f"{PARENT_SPAN_NAME}.{func.__name__}",
                    kind=SpanKind.CLIENT,
                    attributes={SpanAttributes.BASERUN_RUN: Baserun.serialize_run(run)},
                ) as span:
                    try:
                        result = await func(*args, **kwargs)
                        run.result = str(result) if result is not None else ""
                        return result
                    except Exception as e:
                        run.error = "".join(traceback.format_exception(e))
                        raise e
                    finally:
                        Baserun._finish_run(run, span)

        elif inspect.isasyncgenfunction(func):

            async def wrapper(*args, **kwargs):
                if not Baserun._initialized:
                    async for item in func(*args, **kwargs):
                        yield item

                run = Baserun.get_or_create_current_run(
                    name=run_name,
                    trace_type=run_type,
                    metadata=metadata,
                    suite_id=suite_id,
                )

                with tracer.start_as_current_span(
                    f"{PARENT_SPAN_NAME}.{func.__name__}",
                    kind=SpanKind.CLIENT,
                    attributes={SpanAttributes.BASERUN_RUN: Baserun.serialize_run(run)},
                ) as span:
                    try:
                        result = []
                        async for item in func(*args, **kwargs):
                            result.append(item)
                            yield item

                        run.result = str(result) if result is not None else ""
                    except Exception as e:
                        run.error = "".join(traceback.format_exception(e))
                        raise e
                    finally:
                        Baserun._finish_run(run, span)

        else:

            def wrapper(*args, **kwargs):
                if not Baserun._initialized:
                    return func(*args, **kwargs)

                run = Baserun.get_or_create_current_run(
                    name=run_name,
                    trace_type=run_type,
                    metadata=metadata,
                    suite_id=suite_id,
                )

                # Create a parent span so we can attach the run to it, all child spans are part of this run.
                with tracer.start_as_current_span(
                    f"{PARENT_SPAN_NAME}.{func.__name__}",
                    kind=SpanKind.CLIENT,
                    attributes={SpanAttributes.BASERUN_RUN: Baserun.serialize_run(run)},
                ) as span:
                    try:
                        result = func(*args, **kwargs)
                        run.result = str(result) if result is not None else ""
                        return result
                    except Exception as e:
                        run.error = "".join(traceback.format_exception(e))
                        raise e
                    finally:
                        Baserun._finish_run(run, span)

        return wrapper

    @staticmethod
    def trace(func: Callable, metadata: Optional[Dict] = None):
        if Baserun.current_test_suite:
            return Baserun.test(func=func, metadata=metadata)

        return Baserun._trace(
            func=func, run_type=Run.RunType.RUN_TYPE_PRODUCTION, metadata=metadata
        )

    @staticmethod
    def test(func: Callable, metadata: Optional[Dict] = None):
        return Baserun._trace(
            func=func, run_type=Run.RunType.RUN_TYPE_TEST, metadata=metadata
        )

    @staticmethod
    def log(name: str, payload: Union[str, Dict]):
        if not Baserun._initialized:
            return

        run = Baserun.current_run()
        log_message = Log(
            run_id=run.run_id,
            name=name,
            payload=json.dumps(payload),
        )
        log_message.timestamp.FromDatetime(datetime.utcnow())
        log_request = SubmitLogRequest(log=log_message, run=run)

        # noinspection PyBroadException
        try:
            get_or_create_submission_service().SubmitLog(log_request)
        except Exception as e:
            logger.warning(f"Failed to submit log to Baserun: {e}")
